"""
Generate synthetic auxiliary healthcare services dataset
"""

import pandas as pd
import random
from datetime import datetime, timedelta


def generate_reference_period(start_year=2020, end_year=2026):
    """Generate year and month in YYYYMM format"""
    year = random.randint(start_year, end_year)
    month = random.randint(1, 12)
    return f"{year}{month:02d}"


def generate_individual_id():
    """Generate pseudonymised individual ID"""
    return f"IND{random.randint(100000, 999999)}"


def generate_act_id():
    """Generate act identification number"""
    return f"ACT{random.randint(1000000, 9999999)}"


def generate_provider_id():
    """Generate healthcare provider ID"""
    return f"PRV{random.randint(10000, 99999)}"


def generate_auxiliary_service_types():
    """Return list of auxiliary healthcare service types"""
    return [
        "Kinésithérapie",
        "Physiothérapie",
        "Ergothérapie",
        "Orthophonie",
        "Psychothérapie",
        "Diététique",
        "Podologie",
        "Soins infirmiers à domicile",
        "Analyses de laboratoire",
        "Imagerie médicale (radiologie)",
        "Imagerie médicale (IRM)",
        "Imagerie médicale (échographie)",
        "Analyses sanguines",
        "Electrocardiogramme (ECG)",
        "Consultation psychologique",
        "Consultation nutritionniste",
        "Soins de plaies",
        "Perfusion à domicile",
        "Prélèvements biologiques",
        "Oxygénothérapie",
    ]


def generate_service_cost(service_type):
    """Generate realistic cost based on service type"""
    cost_ranges = {
        "Kinésithérapie": (35, 65),
        "Physiothérapie": (40, 70),
        "Ergothérapie": (45, 75),
        "Orthophonie": (50, 80),
        "Psychothérapie": (60, 120),
        "Diététique": (50, 90),
        "Podologie": (40, 70),
        "Soins infirmiers à domicile": (30, 60),
        "Analyses de laboratoire": (20, 150),
        "Imagerie médicale (radiologie)": (50, 200),
        "Imagerie médicale (IRM)": (150, 500),
        "Imagerie médicale (échographie)": (80, 180),
        "Analyses sanguines": (25, 100),
        "Electrocardiogramme (ECG)": (30, 80),
        "Consultation psychologique": (70, 130),
        "Consultation nutritionniste": (55, 95),
        "Soins de plaies": (25, 60),
        "Perfusion à domicile": (45, 120),
        "Prélèvements biologiques": (15, 40),
        "Oxygénothérapie": (80, 200),
    }

    min_cost, max_cost = cost_ranges.get(service_type, (30, 100))
    return round(random.uniform(min_cost, max_cost), 2)


def generate_ss_coverage_ratio(service_type):
    """Generate Social Security coverage ratio (typically 80-100%)"""
    # Most services are covered at 80-88%
    # Some preventive services might have different coverage
    coverage_ratios = [0.80, 0.85, 0.88, 0.90, 0.95, 1.00]
    weights = [30, 25, 20, 15, 7, 3]  # Most common is 80%

    return random.choices(coverage_ratios, weights=weights)[0]


def generate_auxiliary_service_record(individual_ids, provider_ids):
    """Generate one auxiliary service record"""

    reference_period = generate_reference_period(2020, 2025)
    individual_id = random.choice(individual_ids)
    act_id = generate_act_id()

    service_types = generate_auxiliary_service_types()
    auxiliary_service_type = random.choice(service_types)

    provider_id = random.choice(provider_ids)

    # Generate cost
    auxiliary_service_cost = generate_service_cost(auxiliary_service_type)

    # Calculate SS coverage
    coverage_ratio = generate_ss_coverage_ratio(auxiliary_service_type)
    auxiliary_service_cost_paid_ss = round(auxiliary_service_cost * coverage_ratio, 2)

    # Number of invoices (typically 1, but can be multiple for recurring services)
    # Services like physiotherapy often have multiple sessions
    if auxiliary_service_type in [
        "Kinésithérapie",
        "Physiothérapie",
        "Orthophonie",
        "Psychothérapie",
    ]:
        auxiliary_service_nb_invoices = random.choices(
            [1, 2, 3, 4, 5, 6, 8, 10, 12], weights=[20, 15, 12, 10, 8, 10, 8, 10, 7]
        )[0]
    else:
        auxiliary_service_nb_invoices = random.choices([1, 2, 3], weights=[75, 20, 5])[0]

    return {
        "reference_period": reference_period,
        "individual_IDnumber": individual_id,
        "act_IDnumber": act_id,
        "auxiliary_service_type": auxiliary_service_type,
        "provider_IDnumber": provider_id,
        "auxiliary_service_cost": auxiliary_service_cost,
        "auxiliary_service_cost_paid_SS": auxiliary_service_cost_paid_ss,
        "auxiliary_service_nb_invoices": auxiliary_service_nb_invoices,
    }


def generate_dataset(n_individuals=200, n_services=1000, n_providers=50):
    """
    Generate complete auxiliary services dataset

    Parameters:
    - n_individuals: Number of unique individuals (default 200)
    - n_services: Number of service records (default 1000)
    - n_providers: Number of unique healthcare providers (default 50)
    """
    print(f"Generating auxiliary services dataset...")
    print(f"Individuals: {n_individuals}, Services: {n_services}, Providers: {n_providers}")

    # Generate unique individual IDs and provider IDs
    individual_ids = [generate_individual_id() for _ in range(n_individuals)]
    provider_ids = [generate_provider_id() for _ in range(n_providers)]

    # Generate service records
    data = []
    for i in range(n_services):
        if (i + 1) % 200 == 0:
            print(f"Generated {i + 1}/{n_services} service records...")
        data.append(generate_auxiliary_service_record(individual_ids, provider_ids))

    df = pd.DataFrame(data)

    # Sort by reference period and individual ID
    df = df.sort_values(["reference_period", "individual_IDnumber"]).reset_index(drop=True)

    print(f"\nDataset generated successfully!")
    print(f"Total records: {len(df)}")
    print(f"Unique individuals: {df['individual_IDnumber'].nunique()}")
    print(f"Unique providers: {df['provider_IDnumber'].nunique()}")
    print(f"Date range: {df['reference_period'].min()} to {df['reference_period'].max()}")

    return df


def print_dataset_summary(df):
    """Print comprehensive dataset summary"""
    print("\n" + "=" * 80)
    print("DATASET SUMMARY")
    print("=" * 80)

    print(f"\n📊 Basic Statistics:")
    print(f"   Total records: {len(df):,}")
    print(f"   Unique individuals: {df['individual_IDnumber'].nunique():,}")
    print(f"   Unique providers: {df['provider_IDnumber'].nunique():,}")
    print(f"   Period range: {df['reference_period'].min()} - {df['reference_period'].max()}")

    print(f"\n💰 Cost Statistics:")
    print(f"   Total service cost: €{df['auxiliary_service_cost'].sum():,.2f}")
    print(f"   Total SS coverage: €{df['auxiliary_service_cost_paid_SS'].sum():,.2f}")
    print(f"   Average service cost: €{df['auxiliary_service_cost'].mean():.2f}")
    print(f"   Average SS coverage: €{df['auxiliary_service_cost_paid_SS'].mean():.2f}")
    print(
        f"   Average coverage ratio: {(df['auxiliary_service_cost_paid_SS'].sum() / df['auxiliary_service_cost'].sum() * 100):.1f}%"
    )

    print(f"\n🏥 Top 10 Service Types:")
    service_counts = df["auxiliary_service_type"].value_counts().head(10)
    for service, count in service_counts.items():
        pct = (count / len(df)) * 100
        print(f"   {service:.<50} {count:>5} ({pct:>5.1f}%)")

    print(f"\n📅 Services per Period (Top 10):")
    period_counts = df["reference_period"].value_counts().head(10)
    for period, count in period_counts.items():
        print(f"   {period}: {count:>5} services")

    print(f"\n🔢 Invoice Distribution:")
    invoice_dist = df["auxiliary_service_nb_invoices"].value_counts().sort_index()
    for n_invoices, count in invoice_dist.items():
        pct = (count / len(df)) * 100
        print(f"   {int(n_invoices)} invoice(s): {count:>5} ({pct:>5.1f}%)")

    print("\n" + "=" * 80)


if __name__ == "__main__":
    # Generate dataset
    df = generate_dataset(n_individuals=200, n_services=1000, n_providers=50)

    # Save to CSV
    output_file = "data_processed/auxiliary_services_synthetic.csv"
    df.to_csv(output_file, index=False, encoding="utf-8-sig")
    print(f"\n✅ Dataset saved to: {output_file}")

    # Print summary
    print_dataset_summary(df)

    # Show sample records
    print("\n📋 Sample Records:")
    print(df.head(10).to_string(index=False))



from avatars.processors.datetime import DatetimeProcessor
from avatars.processors.expected_mean import ExpectedMeanProcessor
from avatars.processors.geolocation_normalization import (
    GeolocationNormalizationProcessor,
)
from avatars.processors.group_modalities import GroupModalitiesProcessor
from avatars.processors.inter_record_bounded_cumulated_difference import (
    InterRecordBoundedCumulatedDifferenceProcessor,
)
from avatars.processors.inter_record_bounded_range_difference import (
    InterRecordBoundedRangeDifferenceProcessor,
)
from avatars.processors.inter_record_cumulated_difference import (
    InterRecordCumulatedDifferenceProcessor,
)
from avatars.processors.perturbation import PerturbationProcessor
from avatars.processors.proportions import ProportionProcessor
from avatars.processors.relative_difference import RelativeDifferenceProcessor
from avatars.processors.to_categorical import ToCategoricalProcessor

__all__ = [
    "DatetimeProcessor",
    "ExpectedMeanProcessor",
    "GeolocationNormalizationProcessor",
    "GroupModalitiesProcessor",
    "PerturbationProcessor",
    "ProportionProcessor",
    "RelativeDifferenceProcessor",
    "ToCategoricalProcessor",
    "InterRecordCumulatedDifferenceProcessor",
    "InterRecordBoundedRangeDifferenceProcessor",
    "InterRecordBoundedCumulatedDifferenceProcessor",
]



"""
Original source: https://github.com/cgdeboer/iteround

Modified by Olivier Regnier-Coudert (olivier.rc@octopize.io)
Modified by Tom Crasset (tom@octopize.io)
"""

from __future__ import absolute_import, division, print_function

from collections import OrderedDict
from collections.abc import Mapping, Sequence
from enum import Enum
from typing import Any, Callable, Optional, Union, cast


class RoundingStrategy(Enum):
    SMALLEST = "smallest"
    LARGEST = "largest"
    DIFFERENCE = "difference"


class Number(object):
    def __init__(self, order: int, value: Union[float, int]):
        self.value = float(value)
        self.order = order
        self.original = float(value)
        self.diff: float

    def round(self, places: int, rounder: Callable[[float, int], float]) -> None:
        self.value = rounder(self.value, places)
        self.diff = self.original - self.value

    def __repr__(self) -> str:
        return "{} <- {} ({})".format(self.value, self.original, self.diff)


def saferound(
    iterable: Any,
    places: int,
    strategy: RoundingStrategy = RoundingStrategy.DIFFERENCE,
    rounder: Callable[[float, int], float] = round,
    topline: Optional[float] = None,
) -> Any:
    """Rounds an iterable of floats while retaining the original summed value.
    Function parameters should be documented in the ``Args`` section. The name
    of each parameter is required. The type and description of each parameter
    is optional, but should be included if not obvious.
    Args:
        iterable (list, dict, set, numpy.array, generator): list(y) of numbers
            If a dict is passed in, the values must be all floats.
        places (int): Places for rounding.
            Number of places each item in the set should be rounded to.
        strategy (str, optional): The strategy used to clean up rounding errors
            'difference', 'largest', 'smallest'. Defaults to 'difference'
            'difference' seeks to minimize the sum of the array of the
            differences between the original value and the rounded value of
            each item in the iterable. It will adjust the items with the
            largest difference to preserve the sum. This is the default.
            'largest' for any post rounding adjustments, sort the array by
            the largest values to smallest and adjust those first.
            'smallest' for any post rounding adjustments, sort the array by
            the smallest values to largest, adjust the smaller ones first.
        rounder (method, optional): The rounding function. Defaults to `round`
            rounder method must take 2 arguments, a `float` to be rounded, and
            an integer for the number of places (with support for negative)
        topline (float, optional): Topline to match
            Useful in places where we want the total sum to match a different
            topline than the sum of iterable. This can useful in cases where
            original values are altered before passing into the saferound
            method, but the original sum needs to be maintained.
    Returns:
        iterable (same type as input, but with rounded values).
            if 'dict' or 'tuple' are passed, result will be dict or tuple.
            All other iterables (range, map, np.array, etc) will return
            list.
    Raises:
        AssertionError
            - places is not an integer
            - strategy is not valid
            - values are not all floats
    """

    if strategy not in [v for v in RoundingStrategy]:
        raise ValueError(f"Expected a valid RoundingStrategy, got {strategy} instead.")

    values: list[Union[float, int]]

    if isinstance(iterable, Mapping):
        keys, values_view = zip(*iterable.items())
        values = list(values_view)
    elif isinstance(iterable, Sequence):
        values = list(iterable)
    else:
        raise ValueError(f"Expected valid type for iterable, got {type(iterable)} instead.")

    if not all([isinstance(x, float) for x in values]):
        raise ValueError("Expected all values in the iterable to be float.")

    # define a sorting method for rounded differences
    sorter = _sort_by_diff if strategy == RoundingStrategy.DIFFERENCE else _sort_by_value
    default_reverse = False if strategy == RoundingStrategy.SMALLEST else True

    # calculate original sum, rounded,  then rounded local sum.
    local = [Number(i, value) for i, value in enumerate(values)]
    orig_sum = _sumnum(local, places, rounder) if topline is None else rounder(topline, places)

    for n in local:
        n.round(places, rounder)

    local_sum = _sumnum(local, places, rounder)

    # adjust values to adhere to original sum
    while local and local_sum != orig_sum:
        diff = rounder(orig_sum - local_sum, places)
        if diff < 0.0:
            increment = -1 * _mininc(places)
            reverse = False if strategy == RoundingStrategy.DIFFERENCE else default_reverse
        else:
            increment = _mininc(places)
            reverse = True if strategy == RoundingStrategy.DIFFERENCE else default_reverse
        tweaks = int(abs(diff) / _mininc(places))
        local = sorter(local, reverse)
        for ith in range(0, min(tweaks, len(local))):
            local[ith].value += increment
            local[ith].round(places, rounder)
        local_sum = _sumnum(local, places, rounder)

    # return a proper type if passed, else return list
    rounded_list = [n.value for n in _sort_by_order(local)]
    if isinstance(iterable, dict):
        return dict(zip(keys, rounded_list))
    elif isinstance(iterable, OrderedDict):
        return OrderedDict(zip(keys, rounded_list))
    elif isinstance(iterable, tuple):
        return tuple(rounded_list)
    return rounded_list


def _sumnum(numbers: list[Number], places: int, rounder: Callable[[float, int], float]) -> float:
    return rounder(sum([n.value for n in numbers]), places)


def _sort_by_value(numbers: list[Number], reverse: bool) -> list[Number]:
    return sorted(numbers, key=lambda x: x.value, reverse=reverse)


def _sort_by_order(numbers: list[Number]) -> list[Number]:
    return sorted(numbers, key=lambda x: x.order)


def _sort_by_diff(numbers: list[Number], reverse: bool) -> list[Number]:
    return sorted(numbers, key=lambda x: x.diff, reverse=reverse)


def _mininc(places: int) -> float:
    return cast(float, 1 / (10**places))



"""Tests for the crash handler module, specifically sensitive line redaction."""

import textwrap
from pathlib import Path

from avatars import __version__
from avatars.crash_handler import (
    REDACTED_PLACEHOLDER,
    _find_sensitive_lines_from_source,
    _is_sensitive_variable_name,
    generate_crash_report,
    get_full_file_content,
    is_user_code,
)


class TestIsUserCode:
    """Tests for determining if a file is user code or library code."""

    def test_regular_user_file_is_user_code(self) -> None:
        """A regular Python file in a user's project is considered user code."""
        assert is_user_code("/home/user/project/my_script.py") is True
        assert is_user_code("/Users/john/work/analysis.py") is True
        assert is_user_code("C:\\Users\\john\\project\\main.py") is True

    def test_site_packages_is_library_code(self) -> None:
        """Files in site-packages are considered library code."""
        assert is_user_code("/usr/lib/python3.10/site-packages/requests/api.py") is False
        assert is_user_code("/home/user/.venv/lib/python3.10/site-packages/numpy/core.py") is False

    def test_dist_packages_is_library_code(self) -> None:
        """Files in dist-packages are considered library code."""
        assert is_user_code("/usr/lib/python3/dist-packages/apt/cache.py") is False

    def test_lib_python_is_library_code(self) -> None:
        """Files in lib/python paths are considered library code."""
        assert is_user_code("/usr/lib/python3.10/os.py") is False

    def test_avatars_library_is_treated_as_user_code(self) -> None:
        """The avatars library should be treated as user code even when installed."""
        # When avatars is installed in site-packages, it should still show context
        assert (
            is_user_code("/home/user/.venv/lib/python3.10/site-packages/avatars/runner.py") is True
        )
        assert is_user_code("/usr/lib/python3.10/site-packages/avatars/manager.py") is True
        assert (
            is_user_code("/home/user/.venv/lib/python3.10/site-packages/avatars/crash_handler.py")
            is True
        )

    def test_avatars_in_development_is_user_code(self) -> None:
        """The avatars library in development mode is user code."""
        assert is_user_code("/home/user/dev/avatar/services/client/src/avatars/runner.py") is True


class TestIsSensitiveVariableName:
    """Tests for fuzzy matching of sensitive variable names."""

    def test_exact_matches(self) -> None:
        """Exact sensitive names are detected."""
        assert _is_sensitive_variable_name("password") is True
        assert _is_sensitive_variable_name("secret") is True
        assert _is_sensitive_variable_name("apikey") is True
        assert _is_sensitive_variable_name("pwd") is True

    def test_case_insensitive(self) -> None:
        """Matching is case-insensitive."""
        assert _is_sensitive_variable_name("PASSWORD") is True
        assert _is_sensitive_variable_name("Password") is True
        assert _is_sensitive_variable_name("SECRET") is True
        assert _is_sensitive_variable_name("ApiKey") is True

    def test_underscore_variations(self) -> None:
        """Underscores are ignored in matching."""
        assert _is_sensitive_variable_name("api_key") is True
        assert _is_sensitive_variable_name("API_KEY") is True
        assert _is_sensitive_variable_name("pass_word") is True

    def test_common_variations(self) -> None:
        """Common variations of sensitive names are detected."""
        assert _is_sensitive_variable_name("passwd") is True  # 86% match to password
        assert _is_sensitive_variable_name("secrets") is True  # 92% match to secret
        assert _is_sensitive_variable_name("passwords") is True  # high match to password
        assert _is_sensitive_variable_name("api_keys") is True  # high match to apikey

    def test_short_abbreviations_matched(self) -> None:
        """Short abbreviations are matched via explicit patterns."""
        assert _is_sensitive_variable_name("pwd") is True
        assert _is_sensitive_variable_name("PWD") is True
        assert _is_sensitive_variable_name("pw") is True

    def test_non_sensitive_names(self) -> None:
        """Non-sensitive variable names are not flagged."""
        assert _is_sensitive_variable_name("username") is False
        assert _is_sensitive_variable_name("email") is False
        assert _is_sensitive_variable_name("data") is False
        assert _is_sensitive_variable_name("config") is False
        assert _is_sensitive_variable_name("url") is False
        assert _is_sensitive_variable_name("api") is False


class TestFindSensitiveLines:
    """Tests for AST-based sensitive line detection."""

    def test_finds_authenticate_call_with_positional_args(self) -> None:
        """Detect authenticate() call and password variable assignment."""
        source = textwrap.dedent("""
            username = "user@example.com"
            password = "secret123"
            manager.authenticate(username, password)
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # Line 2: password assignment, Line 3: authenticate call
        assert 2 in sensitive_lines
        assert 3 in sensitive_lines
        # Line 1 (username) should not be redacted
        assert 1 not in sensitive_lines

    def test_finds_authenticate_call_with_keyword_args(self) -> None:
        """Detect authenticate() with password= keyword argument."""
        source = textwrap.dedent("""
            my_user = "user"
            my_pass = "secret"
            client.authenticate(username=my_user, password=my_pass)
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # Line 2: password variable assignment, Line 3: authenticate call
        assert 2 in sensitive_lines
        assert 3 in sensitive_lines
        # username variable should not be redacted
        assert 1 not in sensitive_lines

    def test_finds_multiple_authenticate_calls(self) -> None:
        """Detect multiple authenticate() calls with different password vars."""
        source = textwrap.dedent("""
            pw1 = "first_secret"
            manager1.authenticate("user1", pw1)
            pw2 = "second_secret"
            manager2.authenticate("user2", pw2)
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # All password assignments and authenticate calls should be redacted
        assert 1 in sensitive_lines  # pw1 assignment
        assert 2 in sensitive_lines  # first authenticate
        assert 3 in sensitive_lines  # pw2 assignment
        assert 4 in sensitive_lines  # second authenticate

    def test_literal_password_only_redacts_call(self) -> None:
        """When password is a literal, only the authenticate call is redacted."""
        source = textwrap.dedent("""
            username = "user"
            manager.authenticate(username, "literal_password")
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # Only the authenticate call should be redacted (line 2)
        assert 2 in sensitive_lines
        # Username line should not be redacted
        assert 1 not in sensitive_lines

    def test_password_from_env_var(self) -> None:
        """Detect password assignment from os.environ.get()."""
        source = textwrap.dedent("""
            import os
            username = os.environ.get("USERNAME")
            password = os.environ.get("PASSWORD")
            manager.authenticate(username, password)
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # password assignment and authenticate call should be redacted
        assert 3 in sensitive_lines  # password = os.environ.get(...)
        assert 4 in sensitive_lines  # authenticate call
        # import and username should not be redacted
        assert 1 not in sensitive_lines
        assert 2 not in sensitive_lines

    def test_no_authenticate_call(self) -> None:
        """Sensitive variable names are redacted even without authenticate() call."""
        source = textwrap.dedent("""
            password = "secret"
            print(password)
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # password assignment should still be redacted (sensitive variable name)
        assert 1 in sensitive_lines
        # print statement should not be redacted
        assert 2 not in sensitive_lines

    def test_multiline_authenticate_call(self) -> None:
        """Detect multiline authenticate() calls."""
        source = textwrap.dedent("""
            password = "secret"
            manager.authenticate(
                username="user",
                password=password
            )
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # password assignment should be redacted
        assert 1 in sensitive_lines
        # All lines of the multiline call should be redacted
        assert 2 in sensitive_lines
        assert 3 in sensitive_lines
        assert 4 in sensitive_lines
        assert 5 in sensitive_lines

    def test_annotated_assignment(self) -> None:
        """Detect type-annotated password assignments."""
        source = textwrap.dedent("""
            password: str = "secret"
            manager.authenticate("user", password)
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # Both the annotated assignment and authenticate call should be redacted
        assert 1 in sensitive_lines
        assert 2 in sensitive_lines

    def test_syntax_error_returns_empty_set(self) -> None:
        """Invalid Python syntax returns empty set (graceful degradation)."""
        source = "this is not valid python code {"

        sensitive_lines = _find_sensitive_lines_from_source(source)

        assert len(sensitive_lines) == 0

    def test_sensitive_variable_names_without_authenticate(self) -> None:
        """Various sensitive variable names are redacted via fuzzy matching."""
        source = textwrap.dedent("""
            password = "secret1"
            passwd = "secret2"
            secret = "secret3"
            api_key = "key123"
            apikey = "key456"
            secrets = "multiple"
            username = "user"
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # All sensitive variable assignments should be redacted (fuzzy matched)
        assert 1 in sensitive_lines  # password
        assert 2 in sensitive_lines  # passwd (fuzzy matches "password")
        assert 3 in sensitive_lines  # secret
        assert 4 in sensitive_lines  # api_key (underscores removed -> apikey)
        assert 5 in sensitive_lines  # apikey
        assert 6 in sensitive_lines  # secrets (fuzzy matches "secret")
        # username is not sensitive
        assert 7 not in sensitive_lines

    def test_authenticate_on_different_objects(self) -> None:
        """Detect authenticate on any object (client, manager, etc.)."""
        source = textwrap.dedent("""
            secret = "password"
            client.authenticate("user", secret)
            api.authenticate("user", secret)
            auth_client.authenticate("user", secret)
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # All authenticate calls and the secret assignment should be redacted
        assert 1 in sensitive_lines  # secret assignment
        assert 2 in sensitive_lines  # client.authenticate
        assert 3 in sensitive_lines  # api.authenticate
        assert 4 in sensitive_lines  # auth_client.authenticate

    def test_same_variable_reused_for_password(self) -> None:
        """Handle cases where the same variable is assigned multiple times."""
        source = textwrap.dedent("""
            pw = get_password_from_keychain()
            manager.authenticate("user1", pw)
            pw = "new_password"
            manager.authenticate("user2", pw)
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # All pw assignments and authenticate calls should be redacted
        assert 1 in sensitive_lines
        assert 2 in sensitive_lines
        assert 3 in sensitive_lines
        assert 4 in sensitive_lines

    def test_realistic_quickstart_pattern(self) -> None:
        """Test a realistic pattern from the quickstart tutorial."""
        source = textwrap.dedent("""
            import os
            from avatars.manager import Manager

            username = os.environ.get("AVATAR_USERNAME", "")
            password = os.environ.get("AVATAR_PASSWORD", "")

            manager = Manager()
            manager.authenticate(username, password)

            runner = manager.create_runner(set_name="test")
            runner.add_table("data", "data.csv")
        """).strip()

        sensitive_lines = _find_sensitive_lines_from_source(source)

        # password assignment (line 5) and authenticate call (line 8) should be redacted
        assert 5 in sensitive_lines  # password = os.environ.get(...)
        assert 8 in sensitive_lines  # manager.authenticate(...)

        # These should NOT be redacted
        assert 1 not in sensitive_lines  # import os
        assert 2 not in sensitive_lines  # from avatars...
        assert 4 not in sensitive_lines  # username = ...
        assert 7 not in sensitive_lines  # manager = Manager()
        assert 10 not in sensitive_lines  # runner = ...
        assert 11 not in sensitive_lines  # runner.add_table...


class TestCrashReportIntegration:
    """Integration tests verifying the full crash report redacts sensitive info."""

    def test_get_full_file_content_redacts_password(self, tmp_path: Path) -> None:
        """Verify get_full_file_content redacts password lines in a real file."""
        # Create a temporary Python file with sensitive content
        test_file = tmp_path / "user_script.py"
        test_file.write_text(
            textwrap.dedent("""
                import os
                from avatars.manager import Manager

                username = os.environ.get("AVATAR_USERNAME")
                password = os.environ.get("AVATAR_PASSWORD")
                api_key = "super_secret_key_12345"

                manager = Manager()
                manager.authenticate(username, password)

                runner = manager.create_runner(set_name="test")
            """).strip()
        )

        content = get_full_file_content(str(test_file))

        # Verify password-related lines are redacted
        assert REDACTED_PLACEHOLDER in content

        # Verify actual secrets are NOT in the output
        assert "AVATAR_PASSWORD" not in content
        assert "super_secret_key_12345" not in content
        assert "authenticate(username, password)" not in content

        # Verify non-sensitive content IS present
        assert "import os" in content
        assert "AVATAR_USERNAME" in content
        assert 'set_name="test"' in content

    def test_full_crash_report_redacts_sensitive_info(self, tmp_path: Path) -> None:
        """Verify the complete crash report redacts sensitive information in stack context."""
        # Create a temporary Python file with sensitive content that will raise
        test_file = tmp_path / "user_script.py"
        script_content = textwrap.dedent("""
            import os
            password = "my_super_secret_password"
            secret = "another_secret_value"
            username = "safe_username"
            raise ValueError("intentional error for testing")
        """).strip()
        test_file.write_text(script_content)

        # Simulate an exception with a traceback pointing to our test file
        try:
            exec(compile(script_content, str(test_file), "exec"))
        except ValueError:
            import sys

            exc_type, exc_value, exc_tb = sys.exc_info()

            # Generate the crash report
            report = generate_crash_report(exc_type, exc_value, exc_tb)  # type: ignore[arg-type]

            # The stack trace context for the temp file should have redacted lines
            # Find the section for the temp file in the stack trace
            temp_file_section_start = report.find(str(test_file))
            assert temp_file_section_start != -1, "Temp file should be in the report"

            # Extract just the context section for the temp file
            context_start = report.find("Context:", temp_file_section_start)
            # Find the next section (either another [USER]/[LIB] or the === divider)
            next_section = report.find("\n[", context_start)
            full_file_section = report.find("FULL USER FILE CONTENT", context_start)
            end_marker = min(
                next_section if next_section != -1 else len(report),
                full_file_section if full_file_section != -1 else len(report),
            )
            temp_file_context = report[context_start:end_marker]

            # The context for the temp file should have REDACTED placeholders
            assert REDACTED_PLACEHOLDER in temp_file_context

            # The context should NOT show the actual password/secret values
            # (they should be redacted)
            assert "my_super_secret_password" not in temp_file_context
            assert "another_secret_value" not in temp_file_context

            # But safe content should still be visible in context
            assert "safe_username" in temp_file_context

            # The crash report structure should be correct
            assert "CRASH REPORT" in report
            assert "EXCEPTION" in report

    def test_crash_report_includes_sdk_version(self, tmp_path: Path) -> None:
        """Verify the crash report includes the Avatars SDK version."""
        test_file = tmp_path / "simple_script.py"
        script_content = 'raise ValueError("test error")'
        test_file.write_text(script_content)

        try:
            exec(compile(script_content, str(test_file), "exec"))
        except ValueError:
            import sys

            exc_type, exc_value, exc_tb = sys.exc_info()
            report = generate_crash_report(exc_type, exc_value, exc_tb)  # type: ignore[arg-type]

            # Verify the SDK version is in the system information section
            assert f"Avatars SDK: {__version__}" in report



# This file has been generated - DO NOT MODIFY
# API Version : 2.53.1

import warnings
from dataclasses import dataclass
from typing import Optional
from uuid import UUID

import httpx
from structlog import get_logger

from avatars import __version__
from avatars.base_client import BaseClient
from avatars.client_config import ClientConfig
from avatars.constants import DEFAULT_TIMEOUT
from avatars.models import (
    ForgottenPasswordRequest,
    Login,
    LoginResponse,
    ResetPasswordRequest,
)

MAX_FILE_LENGTH = 1024 * 1024 * 1024  # 1 GB

logger = get_logger(__name__)


@dataclass
class AuthTokens:
    access: str
    refresh: Optional[str] = None

    def update(self, resp: LoginResponse) -> None:
        self.access = resp.access_token

        if resp.refresh_token:
            self.refresh = resp.refresh_token


class ApiClient(BaseClient):
    def __init__(
        self,
        base_url: str | None = None,
        *,
        timeout: int | None = None,
        should_verify_ssl: bool | None = None,
        verify_auth: bool = True,
        http_client: Optional[httpx.Client] = None,
        api_key: Optional[str] = None,
        config: ClientConfig | None = None,
    ) -> None:
        """Client to communicate with the Avatar API.

        Parameters
        ----------
        base_url:
            url of the API (required if config not provided)
        timeout:
            timeout in seconds, by default None (uses DEFAULT_TIMEOUT)
        should_verify_ssl : optional
            whether to verify SSL certificates on the server. By default None (uses True)
        http_client : optional
            allow passing in custom httpx.Client instance, by default None
        verify_auth : optional
            Bypass client-side authentication verification, by default True
        api_key : optional
            API key for authentication using api-key-v1 scheme. If provided,
            authenticate() should not be called. By default None
        config : optional
            ClientConfig object containing all configuration.
            Mutually exclusive individual parameters (except http_client).
            By default None
        """
        if config is not None:
            conflicting_params = []
            if base_url is not None:
                conflicting_params.append("base_url")
            if timeout is not None:
                conflicting_params.append("timeout")
            if should_verify_ssl is not None:
                conflicting_params.append("should_verify_ssl")
            if api_key is not None:
                conflicting_params.append("api_key")

            if conflicting_params:
                params_str = ", ".join(conflicting_params)
                raise ValueError(
                    f"Cannot provide both 'config' and other parameters ({params_str}). "
                    "Either pass a ClientConfig object or individual parameters, not both. "
                    "Note: 'http_client' is allowed alongside 'config' for testing."
                )
        else:
            # Create ClientConfig from individual parameters

            if not base_url:
                raise ValueError("base_url must be provided when creating an ApiClient")

            if '"' in base_url:
                raise ValueError(
                    f"Expected base_url not to contain quotes. Got {base_url} instead"
                )

            # Try to derive from base_url (replace /api with /storage)
            if not base_url.endswith("/api"):
                raise ValueError("base_url must end with '/api' to derive storage_endpoint_url")

            final_storage_url = base_url.replace("/api", "/storage")

            config = ClientConfig(
                base_api_url=base_url,
                timeout=timeout,
                should_verify_ssl=(should_verify_ssl if should_verify_ssl is not None else True),
                storage_endpoint_url=final_storage_url,
                api_key=api_key,
            )

        if config.base_api_url is None:
            raise ValueError("base_api_url must be set in ClientConfig")

        final_timeout = config.timeout if config.timeout is not None else DEFAULT_TIMEOUT

        super().__init__(
            base_url=str(config.base_api_url),
            timeout=final_timeout,
            should_verify_ssl=config.should_verify_ssl,
            verify_auth=verify_auth,
            on_auth_refresh=self._refresh_auth if not config.api_key else None,
            http_client=http_client,
            headers={"User-Agent": f"avatar-python/{__version__}"},
            api_key=config.api_key,
        )

        # Importing here to prevent circular import
        from avatars.api import (
            ApiKeys,
            Auth,
            Compatibility,
            Health,
            Jobs,
            Openapi,
            Resources,
            Results,
            Users,
        )
        from avatars.data_upload import DataUploader

        self.api_keys = ApiKeys(self)
        self.auth = Auth(self)
        self.compatibility = Compatibility(self)
        self.health = Health(self)
        self.jobs = Jobs(self)
        self.openapi = Openapi(self)
        self.resources = Resources(self)
        self.results = Results(self)
        self.users = Users(self)

        data_uploader = DataUploader(
            self,
            should_verify_ssl=config.should_verify_ssl,
            storage_endpoint_url=str(config.storage_endpoint_url),
        )
        self.data_uploader = data_uploader
        self.upload_file = data_uploader.upload_file
        self.download_file = data_uploader.download_file

        self.auth_tokens: Optional[AuthTokens] = None

        logger.debug("ApiClient initialized", base_api_url=str(config.base_api_url))

    def authenticate(self, username: str, password: str, timeout: Optional[int] = None) -> None:
        if self._api_key:
            raise ValueError(
                "Cannot call authenticate() when api_key is set. "
                "API key authentication is already active. "
                "To use username/password authentication, create a new ApiClient without api_key."
            )

        resp = self.auth.login(
            Login(username=username, password=password),
            timeout=timeout or self.timeout,
        )
        self._update_auth_tokens(resp)

    def forgotten_password(self, email: str, timeout: Optional[int] = None) -> None:
        self.auth.forgotten_password(
            ForgottenPasswordRequest(email=email), timeout=timeout or self.timeout
        )

    def reset_password(
        self,
        email: str,
        new_password: str,
        new_password_repeated: str,
        token: UUID,
        timeout: Optional[int] = None,
    ) -> None:
        self.auth.reset_password(
            ResetPasswordRequest(
                email=email,
                new_password=new_password,
                new_password_repeated=new_password_repeated,
                token=token,
            ),
            timeout=timeout or self.timeout,
        )

    def __str__(self) -> str:
        return ", ".join(
            f"ApiClient(base_url={self.base_url}"
            f"timeout={self.timeout}"
            f"should_verify_ssl={self.should_verify_ssl}"
            f"verify_auth={self.verify_auth})"
        )

    def _enable_refresh_auth(self, enable: bool = True) -> None:
        self.on_auth_refresh(self._refresh_auth if enable else None)

    def _refresh_auth(self) -> dict[str, str]:
        new_headers: dict[str, str] = {}

        # API keys don't expire, no refresh needed
        if self._api_key:
            return new_headers

        if self.auth_tokens:
            if self.auth_tokens.refresh:
                resp = self.auth.refresh(self.auth_tokens.refresh)
                self._update_auth_tokens(resp, headers=new_headers)
            else:
                warnings.warn("Cannot refresh auth with refresh token")
        else:
            warnings.warn("Client is not authenticated, cannot refresh auth")

        return new_headers

    def _set_auth_bearer(self, token: str, *, headers: Optional[dict[str, str]] = None) -> None:
        self.set_header("Authorization", f"Bearer {token}")

        if headers is not None:
            headers["Authorization"] = f"Bearer {token}"

    def _update_auth_tokens(
        self, resp: LoginResponse, *, headers: Optional[dict[str, str]] = None
    ) -> None:
        if not self.auth_tokens:
            self.auth_tokens = AuthTokens(access=resp.access_token, refresh=resp.refresh_token)
        else:
            self.auth_tokens.update(resp)

        self._set_auth_bearer(self.auth_tokens.access, headers=headers)



import json
from datetime import UTC, datetime, timezone
from pathlib import Path
from typing import Callable, Optional
from unittest.mock import patch
from uuid import UUID, uuid4

import httpx
import numpy as np
import pytest
from polyfactory.factories.pydantic_factory import ModelFactory
from pydantic import BaseModel

from avatars.client import ApiClient
from avatars.log import setup_logging
from avatars.models import (
    BulkDeleteRequest,
    BulkDeleteResponse,
    CompatibilityResponse,
    CompatibilityStatus,
    FileAccess,
    FileCredentials,
    JobCreateRequest,
    JobCreateResponse,
    JobKind,
    JobResponse,
    JobResponseList,
    Login,
    LoginResponse,
    ResourceSetResponse,
)

RequestHandle = Callable[[httpx.Request], httpx.Response]

setup_logging()


def create_fake_job(
    name: str = "test_job",
    set_name: UUID | None = None,
    display_name: str = "test_dataset",
    kind: JobKind = JobKind.standard,
    status: str = "finished",
    parameters_name: str | None = None,
    created_at: datetime | None = None,
    done: bool = True,
    progress: float = 1.0,
    exception: str = "",
) -> JobResponse:
    """Create a fake JobResponse for testing.

    Parameters
    ----------
    name : str
        Job name (default: "test_job")
    set_name : UUID | None
        Set UUID (default: random UUID)
    display_name : str
        Display name for the dataset (default: "test_dataset")
    kind : JobKind
        Job type (default: JobKind.standard)
    status : str
        Job status (default: "finished")
    parameters_name : str | None
        Parameters name (default: matches job kind value)
    created_at : datetime | None
        Creation timestamp (default: now)
    done : bool
        Whether job is complete (default: True)
    progress : float
        Job progress 0-1 (default: 1.0)
    exception : str
        Exception message if any (default: "")

    Returns
    -------
    JobResponse
        A fully populated job object ready for testing

    Examples
    --------
    >>> # Simple usage - all defaults
    >>> job = create_fake_job()

    >>> # Custom set_name and display_name
    >>> job = create_fake_job(set_name=my_uuid, display_name="customers")

    >>> # Job from 5 days ago
    >>> job = create_fake_job(created_at=datetime.now(UTC) - timedelta(days=5))

    >>> # Privacy metrics job
    >>> job = create_fake_job(kind=JobKind.privacy_metrics, name="privacy_job")
    """
    if set_name is None:
        set_name = uuid4()
    if parameters_name is None:
        parameters_name = kind.value
    if created_at is None:
        created_at = datetime.now(UTC)

    return JobResponse(
        name=name,
        set_name=set_name,
        parameters_name=parameters_name,
        created_at=created_at,
        kind=kind,
        status=status,
        exception=exception,
        done=done,
        progress=progress,
        display_name=display_name,
    )


@pytest.fixture(autouse=True)
def no_sleep():
    """Patch time.sleep in avatars.runner to avoid real delays in unit tests."""
    with patch("avatars.runner.time.sleep"):
        yield


def mock_httpx_client(handler: Optional[RequestHandle] = None) -> httpx.Client:
    """Generate a HTTPX client with a MockTransport."""

    if handler is None:
        handler = lambda request: httpx.Response(200, json={})  # noqa: E731

    transport = httpx.MockTransport(handler)
    return httpx.Client(base_url="http://localhost:8000", transport=transport)  # nosec


def api_client_factory(handler: Optional[RequestHandle] = None) -> ApiClient:
    """Generate an API client with a mock transport.

    The handler returns an empty 200 response by default.
    Consider overriding it with a custom handler for more complex tests.
    """
    http_client = mock_httpx_client(handler)
    return ApiClient(
        base_url="http://localhost:8000/api",
        http_client=http_client,
        verify_auth=False,
    )


EXPECTED_KWARGS = ["jobs"]


class FakeJobs:
    def __init__(self, *args, **kwargs):
        kwargs = kwargs or {}
        self._get_jobs_returned_value: JobResponseList | None = kwargs.get(
            "get_jobs_returned_value", None
        )
        self._jobs: dict[str, JobResponse] = {}
        self.bulk_delete_calls: list[BulkDeleteRequest] = []

    def add_job(self, job: JobResponse) -> None:
        """Pre-populate a job into the fake store (for test setup)."""
        self._jobs[job.name] = job

    def create_job(self, request: JobCreateRequest) -> JobCreateResponse:
        name = str(uuid4())
        self._jobs[name] = JobResponse(
            name=name,
            set_name=request.set_name,
            parameters_name=request.parameters_name,
            display_name=request.parameters_name,
            created_at=datetime.now(timezone.utc),
            kind=JobKind.standard,
            status="pending",
            exception="",
            done=False,
        )
        return JobCreateResponse(name=name, Location=f"/jobs/{name}")

    def get_jobs(self) -> JobResponseList:
        if self._get_jobs_returned_value is not None:
            return self._get_jobs_returned_value
        return JobResponseList(jobs=list(self._jobs.values()))

    def get_job_status(self, name: str) -> JobResponse:
        if name in self._jobs:
            return self._jobs[name]
        return JobResponseFactory().build(
            name=name,
            set_name=uuid4(),
            parameters_name=name,
            display_name=name,
            created_at="2023-10-01T00:00:00Z",
            kind=JobKind.standard,
            status="finished",
            done=True,
            progress=1.0,
        )

    def delete_job(self, job_name: str) -> JobResponse:
        if job_name not in self._jobs:
            raise ValueError(f"Job '{job_name}' not found")
        return self._jobs.pop(job_name)

    def bulk_delete_jobs(self, request: BulkDeleteRequest) -> BulkDeleteResponse:
        self.bulk_delete_calls.append(request)
        deleted: list[JobResponse] = []
        failed: list[str] = []
        for job_name in request.job_names:
            if job_name in self._jobs:
                deleted.append(self._jobs.pop(job_name))
            else:
                failed.append(job_name)
        return BulkDeleteResponse(deleted_jobs=deleted, failed_jobs=failed)


def privacy_metrics_factory(table_name: str) -> str:
    return json.dumps(
        {
            "local_cloaking": np.random.uniform(0, 100),
            "hidden_rate": np.random.uniform(0, 100),
            "metadata": {
                "table_name": table_name,
                "computation_type": "standalone",
                "reference": None,
            },
        }
    )


def signal_metrics_factory(table_name: str) -> str:
    return json.dumps(
        {
            "hellinger_mean": np.random.uniform(0, 100),
            "hellinger_std": np.random.uniform(0, 100),
            "metadata": {
                "table_name": table_name,
                "computation_type": "standalone",
                "reference": None,
            },
        }
    )


def run_metadata_factory() -> str:
    return """{
        'created_at': '2025-07-08T12:09:23.820284+00:00',
        'finished_at': '2025-07-08T12:09:24.362366+00:00'
        }"""


def figures_metadata_factory(table_name: str) -> str:
    return json.dumps(
        {
            "kind": "2d_projection",
            "filename": f"{table_name}.2d_projection.html",
            "table_name": table_name,
        }
    )


def figures_factory() -> str:
    return """<div>Test</div>"""


def meta_privacy_metrics_factory() -> str:
    return json.dumps({"meta_privacy": {np.random.uniform(1, 100)}})


def meta_signal_metrics_factory() -> str:
    return json.dumps({"meta_signal": {np.random.uniform(1, 100)}})


def string_table_factory() -> str:
    return """1, 2, 3
    4, 5, 6"""


def advice_factory(table_name: str) -> str:
    return json.dumps(
        {
            "table_name": table_name,
            "advice": {
                "ncp": 30,
                "imputation": {
                    "method": "fast_knn",
                    "k": 3,
                    "training_fraction": 1.0,
                    "return_data_imputed": False,
                },
                "use_categorical_reduction": None,
                "k": 20,
                "summary": "This is a fake advice",
            },
        },
    )


class FakeResults:
    def __init__(self, tables: list[str] | None = None):
        self.tables = tables or []

    def get_permission_to_download(self, url):
        return FileAccess(
            url=url,
            credentials=FileCredentials(
                access_key_id="access_key_id", secret_access_key="secret_access_key"
            ),
        )

    def get_permissions_to_download_batch(self, urls: list[str]):
        """Return FileAccess objects for all URLs in the batch."""
        return [
            FileAccess(
                url=url,
                credentials=FileCredentials(
                    access_key_id="access_key_id", secret_access_key="secret_access_key"
                ),
            )
            for url in urls
        ]

    def get_results(self, job_name):
        results = {}
        for table_name in self.tables:
            if job_name == JobKind.privacy_metrics.value:
                if "privacy_metrics" not in results:
                    results["privacy_metrics"] = []
                results["privacy_metrics"].append(f"fakeurl/{table_name}.privacy.json")
            elif job_name == JobKind.signal_metrics.value:
                if "signal_metrics" not in results:
                    results["signal_metrics"] = []
                results["signal_metrics"].append(f"fakeurl/{table_name}.signal.json")

            elif job_name == JobKind.standard.value:
                if "shuffled" not in results:
                    results["shuffled"] = []
                if "unshuffled" not in results:
                    results["unshuffled"] = []
                if "original_projections" not in results:
                    results["original_projections"] = []
                if "avatar_projections" not in results:
                    results["avatar_projections"] = []
                if "figures" not in results:
                    results["figures"] = []

                results["shuffled"].append(f"fakeurl/{table_name}.shuffled-0.csv")
                results["unshuffled"].append(f"fakeurl/{table_name}.unshuffled-0.csv")
                results["original_projections"].append(
                    f"fakeurl/{table_name}.projections.original-0.csv"
                )
                results["avatar_projections"].append(
                    f"fakeurl/{table_name}.projections.avatars-0.csv"
                )
                results["figures"].append(f"fakeurl/{table_name}.2d_projection.html")
            elif JobKind.advice.value in job_name:
                if "advice" not in results:
                    results["advice"] = []
                results["advice"].append(f"fakeurl/{table_name}.advice.json")

        if job_name == JobKind.privacy_metrics.value:
            results["meta_metrics"] = ["fakeurl/meta_metrics.privacy.json"]
            results["run_metadata"] = ["fakeurl/run_metadata.privacy-metrics.json"]
        elif job_name == JobKind.signal_metrics.value:
            results["meta_metrics"] = ["fakeurl/meta_metrics.signal.json"]
            results["run_metadata"] = ["fakeurl/run_metadata.signal-metrics.json"]
        elif job_name == JobKind.standard.value:
            results["figures_metadata"] = ["fakeurl/figures_metadata.json"]
            results["run_metadata"] = ["fakeurl/run_metadata.avatarize.json"]
        elif job_name == JobKind.report.value:
            results["report"] = ["fakeurl/report.pdf"]
        return results

    def get_upload_url(self):
        raise FileNotFoundError()


class FakeResources:
    def __init__(self):
        self._stored_resources: dict[str, str] = {}

    def put_resources(self, display_name, yaml_string):
        set_name = uuid4()
        self._stored_resources[str(set_name)] = yaml_string
        return ResourceSetResponse(set_name=set_name, display_name=display_name)

    def get_resources(self, set_name: str) -> str:
        """Return mock YAML for a resource set."""
        if set_name in self._stored_resources:
            return self._stored_resources[set_name]
        # Return a minimal valid YAML if not found
        return """---
kind: AvatarMetadata
metadata:
  name: avatar-metadata-test
spec:
  display_name: test_dataset
---
kind: AvatarSchema
metadata:
  name: schema
spec:
  tables:
    - name: table1
      data:
        volume: input
        file: table1.csv
"""


class FakeUser(BaseModel):
    id: UUID


class FakeUsers:
    def __init__(self):
        pass

    def get_me(self):
        return FakeUser(id=uuid4())


class FakeAuth:
    def __init__(self):
        pass

    def login(self, login: Login, timeout: Optional[int] = None):
        return


class FakeCompatibility:
    def __init__(self):
        pass

    def is_client_compatible(self):
        return CompatibilityResponse(
            message="Message from the server",
            status=CompatibilityStatus.incompatible,
            most_recent_compatible_client="1.0.0",
        )


class FakeApiClient(ApiClient):
    """Fake API client for unit testing.

    Provides mock implementations of all API endpoints without requiring
    a real server connection.

    Examples
    --------
    >>> # Simple usage with defaults
    >>> client = FakeApiClient()

    >>> # With pre-configured jobs
    >>> jobs = [create_fake_job(display_name="dataset1")]
    >>> for job in jobs:
    ...     client.jobs.add_job(job)

    >>> # With builder pattern via jobs attribute
    >>> client = FakeApiClient()
    >>> client.jobs.with_job(name="job1").with_job(name="job2")
    """

    jobs: FakeJobs  # type: ignore[assignment]

    def __init__(
        self,
        tables: list[str] | None = None,
    ):
        """Initialize FakeApiClient.

        Parameters
        ----------
        tables : list[str] | None
            List of table names to simulate in results.
        jobs : list[JobWithDisplayNameResponse] | None
            Pre-configured list of jobs for the jobs API.
            Can also be built using client.jobs.with_job().
        """
        self.tables = tables or []
        self.jobs = FakeJobs()  # type: ignore
        self.results = FakeResults(tables=self.tables)  # type: ignore
        self.resources = FakeResources()  # type: ignore
        self.users = FakeUsers()  # type: ignore
        self.base_url = "http://localhost:8000"
        self.auth = FakeAuth()  # type: ignore
        self.compatibility = FakeCompatibility()  # type: ignore
        self.timeout = 100
        # Initialize attributes needed for API key authentication
        self._api_key: Optional[str] = None
        self._headers: dict[str, str] = {}

    def set_header(self, key: str, value: str) -> None:
        """Set a header in the client."""
        self._headers[key] = value

    def _update_auth_tokens(
        self, resp: LoginResponse, *, headers: Optional[dict[str, str]] = None
    ):
        pass

    def upload_file(self, data, key):
        return "File uploaded successfully"

    def download_file(self, file_access):
        if Path(file_access.url).name == "meta_metrics.privacy.json":
            return meta_privacy_metrics_factory()
        elif Path(file_access.url).name == "meta_metrics.signal.json":
            return meta_signal_metrics_factory()
        elif Path(file_access.url).name == "run_metadata.privacy-metrics.json":
            return run_metadata_factory()
        elif Path(file_access.url).name == "run_metadata.signal-metrics.json":
            return run_metadata_factory()
        elif Path(file_access.url).name == "run_metadata.avatarize.json":
            return run_metadata_factory()

        table_name = Path(file_access.url).name.split(".")[0]
        if Path(file_access.url).name.endswith(".html"):
            return figures_factory()
        elif Path(file_access.url).name.endswith("advice.json"):
            return advice_factory(table_name)
        elif Path(file_access.url).name == "report.pdf":
            return b"report content"
        elif Path(file_access.url).name.endswith(".csv"):
            return string_table_factory()
        elif Path(file_access.url).name.endswith(".privacy.json"):
            return privacy_metrics_factory(table_name)
        elif Path(file_access.url).name.endswith(".signal.json"):
            return signal_metrics_factory(table_name)
        elif Path(file_access.url).name == "figures_metadata.json":
            return figures_metadata_factory(self.tables[0])

    def send_request(self, method, url, **kwargs):
        name = kwargs["json_data"].parameters_name
        self.jobs._jobs[name] = JobResponse(
            name=name,
            set_name=kwargs["json_data"].set_name,
            parameters_name=name,
            display_name=name,
            created_at=datetime.now(timezone.utc),
            kind=JobKind.standard,
            status="finished",
            exception="",
            done=True,
        )
        return {"name": name, "Location": ""}


class JobResponseFactory(ModelFactory[JobResponse]):
    __model__ = JobResponse
    __check_model__ = False



import warnings
from typing import List, Optional

import pandas as pd


class ExpectedMeanProcessor:
    """Processor to force values to have similar mean to original data.

    Means and standard deviations are computed for groups of variables and the
    processor ensures that the transformed data has similar mean and std than in
    the original data for each group.
    Care should be taken when using this processor as it only targets enhancement of unimodal
    utility. This may occur at the expense of multi-modal utility and privacy.

    Arguments
    ---------
        target_variables:
            variables to transform

    Keyword Arguments
    -----------------
        groupby_variables:
            variables to use to group values in different distributions
        same_std:
            Set to True to force the variables to transform to have the same
            standard deviation as the reference data. default: False.

    Examples
    --------
    >>> import numpy as np
    >>> df = pd.DataFrame(np.array(([1, 2, 3], [4, 5, 6], [4, 5, 6], [1, 2, 3])),
    ...                   columns=['one', 'two', 'three'])
    >>> df = df.astype('int')
    >>> processor = ExpectedMeanProcessor(target_variables = ['one'])
    >>> processed = processor.preprocess(df)


    The processor forces your synthetic dataset to have the same mean as the original.

    >>> avatar = pd.DataFrame(np.array(([3, 2, 3], [3, 5, 6], [8, 5, 6], [8, 2, 3])),
    ...                   columns=['one', 'two', 'three'])
    >>> avatar.one.mean()
    5.5
    >>> avatar = processor.postprocess(df, avatar)
    >>> avatar.one.mean()
    2.5

    You can also force the mean by category using ```groupby_variables```

    >>> df = pd.DataFrame(
    ...    {
    ...        "variable_1": [11, 24, 23.5, 12],
    ...        "variable_2": ["red", "blue", "blue", "red"],
    ...    }
    ... )
    >>> df
       variable_1 variable_2
    0        11.0        red
    1        24.0       blue
    2        23.5       blue
    3        12.0        red
    >>> df.groupby("variable_2").mean()
    ... # doctest: +NORMALIZE_WHITESPACE
                variable_1
    variable_2
    blue             23.75
    red              11.50
    >>> processor = ExpectedMeanProcessor(
    ...    target_variables = ['variable_1'], groupby_variables= ['variable_2'],
    ...    )
    >>> processor.preprocess(df)
       variable_1 variable_2
    0        11.0        red
    1        24.0       blue
    2        23.5       blue
    3        12.0        red
    >>> avatar = pd.DataFrame(
    ...    {
    ...        "variable_1": [12, 13.5, 23.5, 22],
    ...        "variable_2": ["red", "red", "blue", "blue"],
    ...    }
    ... )
    >>> avatar
       variable_1 variable_2
    0        12.0        red
    1        13.5        red
    2        23.5       blue
    3        22.0       blue
    >>> avatar.groupby("variable_2").mean()
    ... # doctest: +NORMALIZE_WHITESPACE
                variable_1
    variable_2
    blue             22.75
    red              12.75
    >>> avatar = processor.postprocess(df, avatar)
    >>> avatar
       variable_1 variable_2
    0       10.75        red
    1       12.25        red
    2       24.50       blue
    3       23.00       blue
    >>> avatar.groupby("variable_2").mean()
    ... # doctest: +NORMALIZE_WHITESPACE
                variable_1
    variable_2
    blue             23.75
    red              11.50
    """

    def __init__(
        self,
        target_variables: List[str],
        *,
        groupby_variables: Optional[List[str]] = None,
        same_std: bool = False,
    ):
        # Variable added temporarily when no groupby is used. It is a column with only one modality
        # that can be used to perform a groupby in order to get the expected mean and std on all
        # records.
        self.nogroup_name = "___NOGROUPVAR___"
        self.nogroup_value = "__NOGROUPVAL__"
        if groupby_variables is None:
            self.groupby_variables = [self.nogroup_name]
            self.is_nogroup = True
        else:
            self.groupby_variables = groupby_variables
            self.is_nogroup = False
        self.target_variables = target_variables
        self.same_std = same_std

    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        """Compute the reference mean and standard deviations.

        Arguments
        ---------
            df:
                reference dataframe

        Returns
        -------
            df:
                the unaltered reference dataframe.
        """
        working = df.copy()
        if self.is_nogroup:
            working[self.nogroup_name] = [self.nogroup_value for i in range(len(working))]
        cols = self.groupby_variables + self.target_variables
        self.properties_df = _get_distribution_data(
            df=working[cols],
            target_variables=self.target_variables,
            groupby_variables=self.groupby_variables,
        )
        return working

    def postprocess(self, source: pd.DataFrame, dest: pd.DataFrame) -> pd.DataFrame:
        """Force the data to have the reference mean.

        Arguments
        ---------
            source:
                not used
            dest:
                dataframe to transform

        Returns
        -------
            dest:
                a dataframe with the transformed target columns
        """
        original_cols = dest.columns
        if self.is_nogroup:
            # if no groupby, create a one-modality temporary variable
            dest[self.nogroup_name] = [self.nogroup_value] * len(dest)
        cols = self.groupby_variables + self.target_variables
        current_properties_df = _get_distribution_data(
            df=dest[cols],
            target_variables=self.target_variables,
            groupby_variables=self.groupby_variables,
        )
        current_properties_df = current_properties_df.rename(
            columns={f"{col}mean": f"{col}mean_current" for col in self.target_variables}
        )
        current_properties_df = current_properties_df.rename(
            columns={f"{col}std": f"{col}std_current" for col in self.target_variables}
        )

        dest = dest.merge(
            current_properties_df,
            left_on=self.groupby_variables,
            right_on=self.groupby_variables,
        )
        dest = dest.merge(
            self.properties_df,
            left_on=self.groupby_variables,
            right_on=self.groupby_variables,
        )

        for col in self.target_variables:
            same_std_tmp = self.same_std
            if self.same_std and 0 in current_properties_df[f"{col}std_current"].values:
                warnings.warn(
                    f"""target variable {col} has a std of 0. The same standard deviation
                            cannot be guaranteed for this variable.
                    """
                )
                same_std_tmp = False
            if same_std_tmp:
                dest[col] = dest[f"{col}mean"] + (dest[col] - dest[f"{col}mean_current"]) * (
                    dest[f"{col}std"] / dest[f"{col}std_current"]
                )
            else:
                dest[col] = dest[f"{col}mean"] + (dest[col] - dest[f"{col}mean_current"])

        dest = dest[original_cols]

        return dest


def _get_distribution_data(
    df: pd.DataFrame,
    target_variables: List[str],
    groupby_variables: List[str],
) -> pd.DataFrame:
    """Get mean and standard deviation for given variables and groupby.

    Arguments
    ---------
        df:
            data on which to compute statistics
        target_variables:
            list of columns on which to compute statistics
        groupby_variables:
            list of columns to aggregate on

    Returns
    -------
        stats_df:
            statistics dataframe.
    """
    # Perform transformations
    stats_df = df.groupby(groupby_variables, dropna=False).agg(["mean", "std"]).reset_index()

    # Flatten and set new aggregate variable names
    stats_df.columns = pd.Index(["".join(a) for a in stats_df.columns.to_flat_index()])

    return stats_df



"""
Synthetic Absence Dataset Generator

Generates realistic synthetic absence data for 4000 individuals based on the metadata description.
Includes realistic scenarios for different types of absences (sick leave, maternity/paternity,
work accidents, etc.) with appropriate durations and diagnoses.
"""

import pandas as pd
import random
from datetime import datetime, timedelta
from pathlib import Path
import numpy as np

# Seed for reproducibility
random.seed(42)
np.random.seed(42)

DATA_PROCESSED = Path(__file__).parent.parent / "data_processed"

# ============================================================================
# REFERENCE DATA
# ============================================================================

# Absence types for private scheme workers
ABSENCE_TYPES = [
    "Sick Leave",  # Congé de maladie
    "Maternity Leave",  # Congé de maternité
    "Paternity Leave",  # Congé de paternité
    "Work Accident",  # Accident de travail
    "Occupational Disease",  # Maladie professionnelle
    "Long-term Illness",  # Maladie de longue durée
    "Surgery Recovery",  # Convalescence chirurgicale
    "Burnout",  # Épuisement professionnel
    "Mental Health",  # Santé mentale
    "Rehabilitation",  # Rééducation
]

# Common diagnoses by absence type
DIAGNOSES = {
    "Sick Leave": [
        "Influenza",
        "Gastroenteritis",
        "Upper respiratory infection",
        "Bronchitis",
        "Migraine",
        "Back pain",
        "Sinusitis",
        "Pharyngitis",
        "Acute stress reaction",
        "Musculoskeletal disorder",
    ],
    "Maternity Leave": [
        "Normal pregnancy and delivery",
        "Pregnancy complications",
        "Cesarean section",
        "Postpartum recovery",
        "High-risk pregnancy",
    ],
    "Paternity Leave": ["Birth of child", "Adoption"],
    "Work Accident": [
        "Sprain or strain",
        "Fracture",
        "Laceration",
        "Contusion",
        "Burn",
        "Back injury from lifting",
        "Repetitive strain injury",
        "Slip and fall injury",
    ],
    "Occupational Disease": [
        "Carpal tunnel syndrome",
        "Tendinitis",
        "Hearing loss",
        "Dermatitis",
        "Respiratory disease",
        "Asbestosis",
        "Chronic back disorder",
    ],
    "Long-term Illness": [
        "Cancer treatment",
        "Cardiovascular disease",
        "Diabetes complications",
        "Chronic obstructive pulmonary disease",
        "Autoimmune disorder",
        "Neurological disorder",
    ],
    "Surgery Recovery": [
        "Appendectomy",
        "Hernia repair",
        "Knee surgery",
        "Hip replacement",
        "Gallbladder removal",
        "Spinal surgery",
        "Cardiac surgery",
    ],
    "Burnout": [
        "Work-related stress disorder",
        "Chronic fatigue syndrome",
        "Anxiety disorder",
        "Adjustment disorder",
    ],
    "Mental Health": [
        "Depression",
        "Anxiety disorder",
        "Post-traumatic stress disorder",
        "Panic disorder",
        "Bipolar disorder",
    ],
    "Rehabilitation": [
        "Post-surgery rehabilitation",
        "Physical therapy",
        "Occupational therapy",
        "Cardiac rehabilitation",
        "Stroke rehabilitation",
    ],
}

# Typical duration ranges (in days) for each absence type
DURATION_RANGES = {
    "Sick Leave": (1, 14),  # 1-14 days
    "Maternity Leave": (112, 140),  # 16-20 weeks (Luxembourg standard)
    "Paternity Leave": (10, 10),  # 10 days (Luxembourg standard)
    "Work Accident": (3, 90),  # 3 days to 3 months
    "Occupational Disease": (30, 180),  # 1-6 months
    "Long-term Illness": (60, 365),  # 2 months to 1 year
    "Surgery Recovery": (7, 60),  # 1 week to 2 months
    "Burnout": (30, 180),  # 1-6 months
    "Mental Health": (14, 120),  # 2 weeks to 4 months
    "Rehabilitation": (21, 90),  # 3 weeks to 3 months
}

# ============================================================================
# GENERATOR FUNCTIONS
# ============================================================================


def generate_individual_id():
    """Generate pseudonymised 12-character ID"""
    return "ID" + "".join([str(random.randint(0, 9)) for _ in range(10)])


def random_date_in_range(start_date, end_date):
    """Generate random date between start_date and end_date"""
    delta = end_date - start_date
    random_days = random.randint(0, delta.days)
    return start_date + timedelta(days=random_days)


def generate_absence_start_date():
    """Generate absence start date within last 2 years"""
    today = datetime.now()
    two_years_ago = today - timedelta(days=730)
    return random_date_in_range(two_years_ago, today)


def calculate_end_date(start_date, duration_days):
    """Calculate end date based on start date and duration"""
    return start_date + timedelta(days=int(duration_days))


def generate_duration(absence_type):
    """Generate realistic duration based on absence type"""
    min_days, max_days = DURATION_RANGES[absence_type]

    # Use different distributions for different types
    if absence_type == "Sick Leave":
        # Most sick leaves are short (1-3 days), fewer are longer
        weights = [0.4, 0.3, 0.15, 0.08, 0.04, 0.02, 0.01]
        duration_buckets = [1, 2, 3, 5, 7, 10, 14]
        return random.choices(duration_buckets, weights=weights)[0]

    elif absence_type in ["Maternity Leave", "Paternity Leave"]:
        # Maternity/Paternity leave is standardized
        return random.randint(min_days, max_days)

    elif absence_type in ["Long-term Illness", "Burnout", "Mental Health"]:
        # Long-term absences tend toward longer durations
        return int(np.random.triangular(min_days, max_days, max_days))

    else:
        # Other types use triangular distribution (peak in middle)
        return int(np.random.triangular(min_days, (min_days + max_days) / 2, max_days))


def determine_worker_scheme():
    """Determine if individual is private or public scheme worker"""
    # Approximately 75% private, 25% public in Luxembourg
    return random.choices(["private", "public"], weights=[0.75, 0.25])[0]


def should_have_absence(individual_idx, total_individuals):
    """Determine if individual has absence record"""
    # About 60% of workers have at least one absence per year
    # Some individuals may have multiple absences
    if random.random() < 0.40:  # 40% have no absences
        return 0
    elif random.random() < 0.70:  # 70% of those with absences have 1 absence
        return 1
    elif random.random() < 0.85:  # 15% have 2 absences
        return 2
    elif random.random() < 0.95:  # 10% have 3 absences
        return 3
    else:  # 5% have 4+ absences
        return random.randint(4, 6)


def generate_absence_type_weighted():
    """Generate absence type with realistic distribution"""
    types = list(ABSENCE_TYPES)
    # Sick leave is most common
    weights = [
        0.50,  # Sick Leave (most common)
        0.08,  # Maternity Leave
        0.05,  # Paternity Leave
        0.12,  # Work Accident
        0.05,  # Occupational Disease
        0.05,  # Long-term Illness
        0.06,  # Surgery Recovery
        0.04,  # Burnout
        0.03,  # Mental Health
        0.02,  # Rehabilitation
    ]
    return random.choices(types, weights=weights)[0]


def generate_individual_absences(individual_id, num_absences, scheme):
    """Generate absence records for a single individual"""
    absences = []

    if scheme == "public":
        # Public scheme workers: no absence data available
        absences.append(
            {
                "individual_IDnumber": individual_id,
                "a_absence_type": None,
                "a_start_date": None,
                "a_end_date": None,
                "a_length_absence": None,
                "a_diagnosis": None,
            }
        )
    else:
        # Private scheme workers: may have absence records
        if num_absences == 0:
            # No absences recorded
            absences.append(
                {
                    "individual_IDnumber": individual_id,
                    "a_absence_type": None,
                    "a_start_date": None,
                    "a_end_date": None,
                    "a_length_absence": None,
                    "a_diagnosis": None,
                }
            )
        else:
            # Generate multiple absences for this individual
            used_dates = []

            for _ in range(num_absences):
                absence_type = generate_absence_type_weighted()
                duration = generate_duration(absence_type)

                # Generate start date avoiding overlaps
                max_attempts = 20
                for attempt in range(max_attempts):
                    start_date = generate_absence_start_date()
                    end_date = calculate_end_date(start_date, duration)

                    # Check for overlap with existing absences
                    overlap = False
                    for used_start, used_end in used_dates:
                        if start_date <= used_end and end_date >= used_start:
                            overlap = True
                            break

                    if not overlap:
                        used_dates.append((start_date, end_date))
                        break
                else:
                    # If we couldn't find non-overlapping dates, use the last generated ones anyway
                    used_dates.append((start_date, end_date))

                diagnosis = random.choice(DIAGNOSES[absence_type])

                absences.append(
                    {
                        "individual_IDnumber": individual_id,
                        "a_absence_type": absence_type,
                        "a_start_date": start_date.strftime("%Y%m%d"),
                        "a_end_date": end_date.strftime("%Y%m%d"),
                        "a_length_absence": float(duration),
                        "a_diagnosis": diagnosis,
                    }
                )

    return absences


def generate_absence_dataset(num_individuals=4000):
    """
    Generate synthetic absence dataset for specified number of individuals

    Args:
        num_individuals: Number of unique individuals to generate (default: 4000)

    Returns:
        pandas.DataFrame: Dataset with absence records
    """
    print(f"Generating absence dataset for {num_individuals} individuals...")

    all_absences = []

    for i in range(num_individuals):
        if (i + 1) % 500 == 0:
            print(f"  Generated {i + 1}/{num_individuals} individuals...")

        # Generate unique individual ID
        individual_id = generate_individual_id()

        # Determine worker scheme
        scheme = determine_worker_scheme()

        # Determine number of absences for this individual
        num_absences = should_have_absence(i, num_individuals)

        # Generate absence records
        individual_absences = generate_individual_absences(individual_id, num_absences, scheme)
        all_absences.extend(individual_absences)

    # Create DataFrame
    df = pd.DataFrame(all_absences)

    print(f"\nDataset generation complete!")
    print(f"  Total individuals: {num_individuals}")
    print(f"  Total absence records: {len(df)}")
    print(f"  Records with absences: {df['a_absence_type'].notna().sum()}")
    print(f"  Records without absences: {df['a_absence_type'].isna().sum()}")

    return df


def save_dataset(df, filename="absence_dataset.csv"):
    """Save dataset to CSV file"""
    output_path = DATA_PROCESSED / filename
    output_path.parent.mkdir(parents=True, exist_ok=True)
    df.to_csv(output_path, index=False)
    print(f"\nDataset saved to: {output_path}")
    return output_path


def print_dataset_statistics(df):
    """Print statistics about the generated dataset"""
    print("\n" + "=" * 70)
    print("DATASET STATISTICS")
    print("=" * 70)

    # Basic stats
    print(f"\nTotal records: {len(df)}")
    print(f"Unique individuals: {df['individual_IDnumber'].nunique()}")

    # Absence type distribution
    print("\nAbsence Type Distribution:")
    absence_counts = df["a_absence_type"].value_counts()
    for absence_type, count in absence_counts.items():
        if pd.notna(absence_type):
            percentage = (count / len(df[df["a_absence_type"].notna()])) * 100
            print(f"  {absence_type:.<35} {count:>5} ({percentage:>5.1f}%)")

    print(f"\n  {'No absence data':.<35} {df['a_absence_type'].isna().sum():>5}")

    # Duration statistics
    print("\nDuration Statistics (days):")
    print(f"  Mean: {df['a_length_absence'].mean():.1f}")
    print(f"  Median: {df['a_length_absence'].median():.1f}")
    print(f"  Min: {df['a_length_absence'].min():.1f}")
    print(f"  Max: {df['a_length_absence'].max():.1f}")

    # Duration by absence type
    print("\nAverage Duration by Absence Type:")
    for absence_type in df[df["a_absence_type"].notna()]["a_absence_type"].unique():
        avg_duration = df[df["a_absence_type"] == absence_type]["a_length_absence"].mean()
        print(f"  {absence_type:.<35} {avg_duration:>6.1f} days")

    # Most common diagnoses
    print("\nTop 10 Most Common Diagnoses:")
    top_diagnoses = df["a_diagnosis"].value_counts().head(10)
    for diagnosis, count in top_diagnoses.items():
        if pd.notna(diagnosis):
            print(f"  {diagnosis:.<50} {count:>4}")

    print("\n" + "=" * 70)


# ============================================================================
# MAIN EXECUTION
# ============================================================================

if __name__ == "__main__":
    # Generate dataset
    df = generate_absence_dataset(num_individuals=4000)

    # Print statistics
    print_dataset_statistics(df)

    # Save to CSV
    save_dataset(df, "absence_dataset.csv")

    # Display sample records
    print("\nSample Records (first 10 with absences):")
    print(df[df["a_absence_type"].notna()].head(10).to_string())

    print("\nSample Records (without absences):")
    print(df[df["a_absence_type"].isna()].head(5).to_string())



import logging.config
import os

import structlog

LOG_LEVEL = str(os.getenv("LOG_LEVEL", "WARNING")).upper()


def setup_logging(log_level=LOG_LEVEL):
    timestamper = structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S")
    structlog.configure(
        processors=[
            structlog.stdlib.add_log_level,
            structlog.stdlib.add_logger_name,
            structlog.stdlib.PositionalArgumentsFormatter(),
            timestamper,
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
            # add logger name to the event_dict
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    pre_chain = [
        # Add the log level and a timestamp to the event_dict if the log entry
        # is not from structlog.
        structlog.stdlib.add_log_level,
        # Add extra attributes of LogRecord objects to the event dictionary
        # so that values passed in the extra parameter of log methods pass
        # through to log output.
        structlog.stdlib.ExtraAdder(),
        timestamper,
    ]

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "plain": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": [
                        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                        structlog.dev.ConsoleRenderer(colors=False),
                    ],
                    "foreign_pre_chain": pre_chain,
                },
                "colored": {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processors": [
                        structlog.stdlib.ProcessorFormatter.remove_processors_meta,
                        structlog.dev.ConsoleRenderer(colors=True),
                    ],
                    "foreign_pre_chain": pre_chain,
                },
            },
            "handlers": {
                "default": {
                    "level": log_level,
                    "class": "logging.StreamHandler",
                    "formatter": "colored",
                }
            },
            "loggers": {
                "": {
                    "handlers": ["default"],
                    "level": log_level,
                    "propagate": True,
                },
            },
        }
    )



"""
Synthetic Job Characteristics Dataset Generator

Generates realistic synthetic job/employment data for 4000 individuals based on Luxembourg's
labor market metadata. Includes realistic scenarios for different employment types, wages,
working hours, and contract types.

Author: Data Generator
Date: January 2026
"""

import pandas as pd
import random
from datetime import datetime, timedelta
from pathlib import Path
import numpy as np

# Seed for reproducibility
random.seed(42)
np.random.seed(42)

DATA_PROCESSED = Path(__file__).parent.parent / "data_processed"

# ============================================================================
# REFERENCE DATA
# ============================================================================

# Job sectors in Luxembourg
JOB_SECTORS = {
    "Finance & Insurance": [
        "Bank Analyst",
        "Financial Advisor",
        "Insurance Broker",
        "Accountant",
        "Auditor",
    ],
    "IT & Technology": [
        "Software Developer",
        "Data Analyst",
        "IT Consultant",
        "System Administrator",
        "DevOps Engineer",
    ],
    "Healthcare": ["Nurse", "Doctor", "Medical Assistant", "Pharmacist", "Physiotherapist"],
    "Education": ["Teacher", "Professor", "Education Assistant", "School Administrator", "Tutor"],
    "Construction": ["Construction Worker", "Electrician", "Plumber", "Carpenter", "Site Manager"],
    "Retail & Commerce": ["Sales Associate", "Store Manager", "Cashier", "Buyer", "Merchandiser"],
    "Hospitality": ["Chef", "Waiter", "Hotel Manager", "Receptionist", "Housekeeper"],
    "Manufacturing": [
        "Machine Operator",
        "Quality Controller",
        "Production Manager",
        "Warehouse Worker",
        "Technician",
    ],
    "Transportation": [
        "Truck Driver",
        "Bus Driver",
        "Logistics Coordinator",
        "Delivery Driver",
        "Dispatcher",
    ],
    "Public Administration": [
        "Administrative Assistant",
        "Civil Servant",
        "Policy Analyst",
        "Urban Planner",
        "Social Worker",
    ],
    "Agriculture": [
        "Farmer",
        "Agricultural Worker",
        "Vineyard Worker",
        "Farm Manager",
        "Agricultural Technician",
    ],
}

# Contract types
CONTRACT_TYPES = {
    "0": "Permanent contract",
    "1": "Fixed-term contract",
    "2": "Temporary contract",
    "3": "Apprenticeship contract",
    "4": "Summer or student job",
}

# Job status types
JOB_STATUS = {"0": "Private sector worker", "1": "Civil servant", "2": "Self-employed"}

# NACE sectors (simplified)
NACE_SECTORS = [
    "K64",
    "K65",
    "K66",  # Financial services
    "J62",
    "J63",  # IT services
    "Q86",
    "Q87",  # Healthcare
    "P85",  # Education
    "F41",
    "F42",
    "F43",  # Construction
    "G47",  # Retail
    "I56",  # Food services
    "C10",
    "C20",
    "C30",  # Manufacturing
    "H49",
    "H52",  # Transportation
    "O84",  # Public administration
    "A01",
    "A02",  # Agriculture
]

# Minimum wage in Luxembourg (2026 est.)
MINIMUM_WAGE_HOURLY = 15.50  # EUR
NORMAL_MONTHLY_HOURS = 173  # Standard monthly hours

# ============================================================================
# GENERATOR FUNCTIONS
# ============================================================================


def generate_id(prefix, length=10):
    """Generate pseudonymised ID"""
    return prefix + "".join([str(random.randint(0, 9)) for _ in range(length)])


def generate_reference_period():
    """Generate year-month reference period"""
    year = random.choice([2024, 2025, 2026])
    month = random.randint(1, 12)
    return f"{year}{month:02d}"


def random_date_yyyymm(start_year=2015, end_year=2026):
    """Generate random date in YYYYMM format"""
    year = random.randint(start_year, end_year)
    month = random.randint(1, 12)
    return f"{year}{month:02d}"


def select_job_status():
    """Select job status with realistic distribution"""
    # 70% private, 20% civil servant, 10% self-employed
    return random.choices(list(JOB_STATUS.keys()), weights=[0.70, 0.20, 0.10])[0]


def select_contract_type(job_status):
    """Select contract type based on job status"""
    if job_status == "1":  # Civil servant - permanent
        return "0"
    elif job_status == "2":  # Self-employed - not applicable
        return "-9"
    else:  # Private sector
        # 70% permanent, 20% fixed-term, 5% temporary, 3% apprentice, 2% student
        return random.choices(["0", "1", "2", "3", "4"], weights=[0.70, 0.20, 0.05, 0.03, 0.02])[0]


def generate_job_duration(contract_type):
    """Generate job start and end dates based on contract type"""
    start_year = random.randint(2015, 2025)
    start_month = random.randint(1, 11)
    start_date = f"{start_year}{start_month:02d}"

    if contract_type == "0":  # Permanent - no end date or future end
        if random.random() < 0.85:  # 85% still active
            end_date = None
            planned_end = "999909"
            planned_duration = "-9"
        else:  # 15% ended
            end_year = random.randint(start_year, 2026)
            if end_year == start_year:
                end_month = random.randint(start_month + 1, 12)
            else:
                end_month = random.randint(1, 12)
            end_date = f"{end_year}{end_month:02d}"
            planned_end = "999909"
            planned_duration = "-9"

    elif contract_type == "1":  # Fixed-term
        duration_category = random.choices(
            ["0", "1", "2", "3", "4", "5", "6"], weights=[0.05, 0.10, 0.15, 0.30, 0.25, 0.10, 0.05]
        )[0]

        duration_months = {"0": 0.5, "1": 1, "2": 1.5, "3": 4, "4": 9, "5": 18, "6": 30}[
            duration_category
        ]

        end_year = start_year
        end_month = start_month + int(duration_months)
        while end_month > 12:
            end_month -= 12
            end_year += 1

        if end_year > 2026 or (end_year == 2026 and end_month > 1):
            end_date = None  # Still active
        else:
            end_date = f"{end_year}{end_month:02d}"

        planned_end = f"{end_year}{end_month:02d}"
        planned_duration = duration_category

    elif contract_type == "3":  # Apprenticeship - typically 2-3 years
        duration_months = random.choice([24, 36])
        end_year = start_year + (duration_months // 12)
        end_month = start_month

        if end_year > 2026:
            end_date = None
        else:
            end_date = f"{end_year}{end_month:02d}"

        planned_end = f"{end_year}{end_month:02d}"
        planned_duration = "5"  # 12-24 months category

    elif contract_type == "4":  # Summer/student job - 1-3 months
        duration_months = random.randint(1, 3)
        end_year = start_year
        end_month = start_month + duration_months
        if end_month > 12:
            end_month -= 12
            end_year += 1

        end_date = f"{end_year}{end_month:02d}"
        planned_end = end_date
        planned_duration = random.choice(["0", "1", "2"])

    else:  # Temporary or other
        duration_months = random.randint(1, 6)
        end_year = start_year
        end_month = start_month + duration_months
        if end_month > 12:
            end_month -= 12
            end_year += 1

        if end_year > 2026:
            end_date = None
        else:
            end_date = f"{end_year}{end_month:02d}"

        planned_end = f"{end_year}{end_month:02d}"
        planned_duration = random.choice(["2", "3", "4"])

    return start_date, end_date, planned_end, planned_duration


def is_blue_collar(sector):
    """Determine if job is blue collar based on sector"""
    blue_collar_sectors = [
        "Construction",
        "Manufacturing",
        "Transportation",
        "Agriculture",
        "Hospitality",
    ]
    return "1" if sector in blue_collar_sectors else "0"


def generate_wage(job_status, is_blue_collar_worker, hours_worked):
    """Generate realistic wages based on job characteristics"""
    if job_status == "2":  # Self-employed - not applicable
        return None, None, None, None, None

    # Base hourly wage varies by sector and skill level
    if job_status == "1":  # Civil servant - higher wages
        hourly_base = random.uniform(22, 45)
    elif is_blue_collar_worker == "1":
        hourly_base = random.uniform(15.5, 28)
    else:  # White collar
        hourly_base = random.uniform(18, 50)

    # Calculate monthly base wage
    base_wage = hourly_base * hours_worked if hours_worked else hourly_base * NORMAL_MONTHLY_HOURS

    # Additional wage (bonuses, commissions) - 30% chance
    if random.random() < 0.30:
        additional_wage = random.uniform(100, 1000)
    else:
        additional_wage = 0

    # Overtime - 15% chance
    if random.random() < 0.15 and job_status != "1":
        overtime_hours = random.uniform(5, 20)
        overtime_wage = hourly_base * 1.5 * overtime_hours
    else:
        overtime_hours = 0
        overtime_wage = 0

    total_wage = base_wage + additional_wage + overtime_wage

    return (
        round(base_wage, 2),
        round(additional_wage, 2),
        round(overtime_wage, 2),
        round(total_wage, 2),
        overtime_hours,
    )


def generate_working_hours(job_status, contract_type):
    """Generate realistic working hours"""
    if job_status == "2":  # Self-employed - not applicable
        return None, None, None, None, None

    # Base hours vary by contract type
    if contract_type == "4":  # Student job - part time
        base_hours = random.uniform(60, 120)
    elif contract_type == "3":  # Apprenticeship
        base_hours = random.uniform(140, 173)
    else:  # Full time
        base_hours = random.uniform(160, 173)

    # Overtime - 15% of workers
    if random.random() < 0.15:
        overtime_hours = random.uniform(5, 20)
    else:
        overtime_hours = 0

    total_hours = base_hours + overtime_hours

    # Adjusted hours (calendar effect)
    base_hours_adj = base_hours * random.uniform(0.98, 1.02)
    total_hours_adj = total_hours * random.uniform(0.98, 1.02)

    return (
        round(base_hours, 2),
        round(overtime_hours, 2),
        round(total_hours, 2),
        round(base_hours_adj, 2),
        round(total_hours_adj, 2),
    )


def generate_trial_period(contract_type, start_date):
    """Generate trial period information"""
    if contract_type in ["0", "1"]:  # Permanent or fixed-term - 70% have trial period
        if random.random() < 0.70:
            trial_weeks = random.choice([4, 6, 8, 12, 26])  # Common trial periods

            # Calculate end date
            start_year = int(start_date[:4])
            start_month = int(start_date[4:6])
            trial_months = trial_weeks // 4
            end_month = start_month + trial_months
            end_year = start_year
            if end_month > 12:
                end_month -= 12
                end_year += 1

            trial_end = f"{end_year}{end_month:02d}"
            return "1", trial_end, float(trial_weeks)
        else:
            return "0", None, None
    else:
        return "-9", None, None


def generate_special_statuses():
    """Generate special employment statuses"""
    # Posted worker - 2% of private sector
    posted = "1" if random.random() < 0.02 else "0"

    # Assisting family member - 1% (mainly agriculture/commerce)
    assisting_family = "1" if random.random() < 0.01 else "0"

    # Maternity leave - 3% of workers at any time
    maternity_leave = "1" if random.random() < 0.03 else "0"

    # Activation measure (ADEM program) - 5%
    activation = "1" if random.random() < 0.05 else "0"
    activation_type = (
        random.choice(
            [
                "CAE - Contrat d'appui-emploi",
                "CIE - Contrat d'initiation à l'emploi",
                "PAN - Projet d'activité national",
            ]
        )
        if activation == "1"
        else None
    )

    # Short-time working (chômage partiel) - 8%
    short_time = "1" if random.random() < 0.08 else "0"

    return posted, assisting_family, maternity_leave, activation, activation_type, short_time


def generate_job_record(individual_id):
    """Generate a complete job record for an individual"""
    job_id = generate_id("JOB", 12)
    employer_id = generate_id("EMP", 10)
    reference_period = generate_reference_period()

    # Job status determines many other fields
    job_status = select_job_status()
    contract_type = select_contract_type(job_status)

    # Generate dates
    start_date, end_date, planned_end, planned_duration = generate_job_duration(contract_type)

    # Sector and blue collar status
    sector = random.choice(list(JOB_SECTORS.keys()))
    blue_collar = is_blue_collar(sector) if job_status != "2" else "-8"

    # Working hours
    base_hours, overtime_hours, total_hours, base_hours_adj, total_hours_adj = (
        generate_working_hours(job_status, contract_type)
    )

    # Wages
    if base_hours:
        base_wage, additional_wage, overtime_wage, total_wage, ot_hours = generate_wage(
            job_status, blue_collar, base_hours
        )
        if ot_hours:
            overtime_hours = ot_hours
            total_hours = base_hours + overtime_hours
            total_hours_adj = base_hours_adj + overtime_hours
    else:
        base_wage = additional_wage = overtime_wage = total_wage = None

    # Calculate hourly wages
    if base_hours_adj and base_wage:
        hourly_base = round(base_wage / base_hours_adj, 2)
    else:
        hourly_base = None

    if total_hours_adj and total_wage:
        hourly_total = round(total_wage / total_hours_adj, 2)
    else:
        hourly_total = None

    # Trial period
    trial_period, trial_end, trial_weeks = generate_trial_period(contract_type, start_date)

    # Special statuses
    posted, assisting_family, maternity_leave, activation, activation_type, short_time = (
        generate_special_statuses()
    )

    # Short-time working details
    if short_time == "1" and base_wage and base_hours:
        short_time_benefit = round(base_wage * 0.3, 2)  # 30% of wage
        short_time_hours = round(base_hours * 0.3, 2)  # 30% of hours
    else:
        short_time_benefit = None
        short_time_hours = None

    # Temporary jobs employer (5% of cases)
    temp_employer = generate_id("TMP", 10) if random.random() < 0.05 else "-9"

    # Official statistics inclusion (95% included)
    official_stat = "1" if random.random() < 0.95 else "0"

    # Build record
    record = {
        "reference_period": reference_period,
        "job_IDnumber": job_id,
        "individual_IDnumber": individual_id,
        "employer_IDnumber": employer_id if job_status != "2" else "-9",
        "temporary_jobs_employer_IDnumber": temp_employer,
        "j_job_start_date": start_date,
        "j_job_end_effective_date": end_date if end_date else None,
        "j_contract_type": contract_type,
        "j_job_end_planned_date": planned_end,
        "j_contract_planned_duration": planned_duration,
        "j_job_status": job_status,
        "j_blue_collar": blue_collar,
        "j_posted_worker": posted if job_status != "2" else "-9",
        "j_assisting_family_member": assisting_family,
        "j_maternity_leave": maternity_leave,
        "j_activation_measure": activation,
        "j_activation_measure_type": activation_type,
        "j_trial_period": trial_period,
        "j_trial_period_end_date": trial_end,
        "j_trial_period_length": trial_weeks,
        "j_F_short_time_working": short_time if job_status != "2" else "-9",
        "j_short_time_working_benefit": short_time_benefit,
        "j_short_time_working_hours": short_time_hours,
        "j_F_month_base_wage": "1" if base_wage else ("-9" if job_status == "2" else "-8"),
        "j_month_base_wage": base_wage,
        "j_F_month_additionnal_wage": "1"
        if additional_wage and additional_wage > 0
        else ("-9" if job_status == "2" else "-8"),
        "j_month_additionnal_wage": additional_wage
        if additional_wage and additional_wage > 0
        else None,
        "j_F_month_overtime_wage": "1"
        if overtime_wage and overtime_wage > 0
        else ("-9" if job_status == "2" else "-8"),
        "j_month_overtime_wage": overtime_wage if overtime_wage and overtime_wage > 0 else None,
        "j_F_month_total_wage": "1" if total_wage else ("-9" if job_status == "2" else "-8"),
        "j_month_total_wage": total_wage,
        "j_F_nb_base_worked_hours": "1" if base_hours else ("-9" if job_status == "2" else "-8"),
        "j_nb_base_worked_hours": base_hours,
        "j_F_nb_overtime_paid_hours": "1"
        if overtime_hours and overtime_hours > 0
        else ("-9" if job_status == "2" else "-8"),
        "j_nb_overtime_paid_hours": overtime_hours
        if overtime_hours and overtime_hours > 0
        else None,
        "j_F_nb_total_worked_hours": "1" if total_hours else ("-9" if job_status == "2" else "-8"),
        "j_nb_total_worked_hours": total_hours,
        "j_F_nb_base_worked_hours_ADJ": "1"
        if base_hours_adj
        else ("-9" if job_status == "2" else "-8"),
        "j_nb_base_worked_hours_ADJ": base_hours_adj,
        "j_F_nb_total_worked_hours_ADJ": "1"
        if total_hours_adj
        else ("-9" if job_status == "2" else "-8"),
        "j_nb_total_worked_hours_ADJ": total_hours_adj,
        "j_F_hourly_base_wage": "1" if hourly_base else ("-9" if job_status == "2" else "-8"),
        "j_hourly_base_wage": hourly_base,
        "j_F_hourly_total_wage": "1" if hourly_total else ("-9" if job_status == "2" else "-8"),
        "j_hourly_total_wage": hourly_total,
        "j_F_official_stat": official_stat,
    }

    return record


def generate_job_dataset(num_individuals=4000):
    """
    Generate synthetic job characteristics dataset

    Args:
        num_individuals: Number of unique individuals (default: 4000)

    Returns:
        pandas.DataFrame: Dataset with job records
    """
    print(f"Generating job characteristics dataset for {num_individuals} individuals...")

    all_jobs = []

    for i in range(num_individuals):
        if (i + 1) % 500 == 0:
            print(f"  Generated {i + 1}/{num_individuals} individuals...")

        # Generate unique individual ID
        individual_id = generate_id("ID", 10)

        # Some individuals may have multiple jobs (10%)
        num_jobs = random.choices([1, 2], weights=[0.90, 0.10])[0]

        for _ in range(num_jobs):
            job_record = generate_job_record(individual_id)
            all_jobs.append(job_record)

    # Create DataFrame
    df = pd.DataFrame(all_jobs)

    print(f"\nDataset generation complete!")
    print(f"  Total individuals: {num_individuals}")
    print(f"  Total job records: {len(df)}")

    return df


def save_dataset(df, filename="job_characteristics_dataset.csv"):
    """Save dataset to CSV file"""
    output_path = DATA_PROCESSED / filename
    output_path.parent.mkdir(parents=True, exist_ok=True)
    df.to_csv(output_path, index=False)
    print(f"\nDataset saved to: {output_path}")
    return output_path


def print_dataset_statistics(df):
    """Print statistics about the generated dataset"""
    print("\n" + "=" * 70)
    print("JOB CHARACTERISTICS DATASET STATISTICS")
    print("=" * 70)

    print(f"\nTotal job records: {len(df)}")
    print(f"Unique individuals: {df['individual_IDnumber'].nunique()}")
    print(
        f"Unique employers: {df[df['employer_IDnumber'] != '-9']['employer_IDnumber'].nunique()}"
    )

    # Job status distribution
    print("\nJob Status Distribution:")
    for status_code, status_name in JOB_STATUS.items():
        count = (df["j_job_status"] == status_code).sum()
        percentage = (count / len(df)) * 100
        print(f"  {status_name:.<35} {count:>5} ({percentage:>5.1f}%)")

    # Contract type distribution
    print("\nContract Type Distribution:")
    for contract_code, contract_name in CONTRACT_TYPES.items():
        count = (df["j_contract_type"] == contract_code).sum()
        percentage = (
            (count / len(df[df["j_contract_type"] != "-9"])) * 100
            if len(df[df["j_contract_type"] != "-9"]) > 0
            else 0
        )
        print(f"  {contract_name:.<35} {count:>5} ({percentage:>5.1f}%)")

    # Wage statistics (private sector only)
    private_sector = df[df["j_month_total_wage"].notna()]
    if len(private_sector) > 0:
        print("\nMonthly Wage Statistics (EUR):")
        print(f"  Mean: {private_sector['j_month_total_wage'].mean():,.2f}")
        print(f"  Median: {private_sector['j_month_total_wage'].median():,.2f}")
        print(f"  Min: {private_sector['j_month_total_wage'].min():,.2f}")
        print(f"  Max: {private_sector['j_month_total_wage'].max():,.2f}")

        print("\nHourly Wage Statistics (EUR):")
        hourly_data = df[df["j_hourly_total_wage"].notna()]
        if len(hourly_data) > 0:
            print(f"  Mean: {hourly_data['j_hourly_total_wage'].mean():.2f}")
            print(f"  Median: {hourly_data['j_hourly_total_wage'].median():.2f}")
            print(f"  Min: {hourly_data['j_hourly_total_wage'].min():.2f}")
            print(f"  Max: {hourly_data['j_hourly_total_wage'].max():.2f}")

    # Working hours statistics
    hours_data = df[df["j_nb_total_worked_hours"].notna()]
    if len(hours_data) > 0:
        print("\nMonthly Working Hours Statistics:")
        print(f"  Mean: {hours_data['j_nb_total_worked_hours'].mean():.1f}")
        print(f"  Median: {hours_data['j_nb_total_worked_hours'].median():.1f}")
        print(f"  Min: {hours_data['j_nb_total_worked_hours'].min():.1f}")
        print(f"  Max: {hours_data['j_nb_total_worked_hours'].max():.1f}")

    # Special statuses
    print("\nSpecial Employment Statuses:")
    print(f"  Blue collar workers: {(df['j_blue_collar'] == '1').sum()}")
    print(f"  Posted workers: {(df['j_posted_worker'] == '1').sum()}")
    print(f"  Maternity leave: {(df['j_maternity_leave'] == '1').sum()}")
    print(f"  Activation measures: {(df['j_activation_measure'] == '1').sum()}")
    print(f"  Short-time working: {(df['j_F_short_time_working'] == '1').sum()}")
    print(f"  Trial period: {(df['j_trial_period'] == '1').sum()}")

    print("\n" + "=" * 70)


# ============================================================================
# MAIN EXECUTION
# ============================================================================

if __name__ == "__main__":
    # Generate dataset
    df = generate_job_dataset(num_individuals=4000)

    # Print statistics
    print_dataset_statistics(df)

    # Save to CSV
    save_dataset(df, "job_characteristics_dataset.csv")

    # Display sample records
    print("\nSample Records (first 5):")
    sample_cols = [
        "individual_IDnumber",
        "j_job_status",
        "j_contract_type",
        "j_month_total_wage",
        "j_nb_total_worked_hours",
        "j_hourly_total_wage",
    ]
    print(df[sample_cols].head(10).to_string())



import os
from datetime import UTC, datetime, timedelta
from unittest import mock
from uuid import uuid4

import pytest

from avatars.client import ApiClient
from avatars.client_config import ClientConfig
from avatars.config import Config
from avatars.manager import Manager, Runner, _increment_display_name_version
from avatars.models import (
    BulkDeleteRequest,
    BulkDeleteResponse,
    JobCreateRequest,
    JobKind,
)
from tests.unit.conftest import (
    FakeApiClient,
    JobResponseFactory,
    create_fake_job,
)


class TestManager:
    manager: Manager

    @classmethod
    def setup_class(cls):
        api_client = FakeApiClient()
        cls.manager = Manager(
            api_client=api_client,  # type: ignore[arg-type]
        )

    def test_get_last_job(self) -> None:
        api_client = FakeApiClient()
        for job in JobResponseFactory.batch(2):
            api_client.jobs.add_job(job)
        manager = Manager(
            api_client=api_client,  # type: ignore[arg-type]
        )
        results = manager.get_last_results(1)  # check the get result mock
        assert len(results) == 1

    def test_create_runner(self) -> None:
        runner = self.manager.create_runner("test")
        assert runner is not None
        assert isinstance(runner, Runner)

    @pytest.mark.parametrize(
        "incompatibility_status",
        [
            "incompatible",
            "unknown",
        ],
    )
    def test_should_verify_compatibility(self, incompatibility_status: str) -> None:
        """Verify that the client raises a DeprecationWarning when the server is incompatible"""
        with pytest.raises(DeprecationWarning, match="Client is not compatible with the server."):
            with pytest.warns(DeprecationWarning):
                self.manager.authenticate(
                    username="username",
                    password="password",
                    should_verify_compatibility=True,
                )

    def test_should_not_verify_compatibility(self) -> None:
        """Verify that the client does not raise when should_verify_compatibility is False"""
        try:
            self.manager.authenticate(
                username="username",
                password="password",
                should_verify_compatibility=False,
            )
        except DeprecationWarning:
            pytest.fail("DeprecationWarning was raised unexpectedly.")

    def test_verify_compatibility_uses_config_default(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Verify manager uses config default when should_verify_compatibility is not provided."""

        # Patch config in the manager module where it's imported
        test_config = Config(VERIFY_COMPATIBILITY=True)
        monkeypatch.setattr("avatars.manager.config", test_config)

        # When should_verify_compatibility is not provided, it should use config and raise
        with pytest.raises(DeprecationWarning, match="Client is not compatible with the server."):
            with pytest.warns(DeprecationWarning):
                self.manager.authenticate(
                    username="username",
                    password="password",
                )

    def test_verify_compatibility_uses_config_false(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Verify manager uses config value when set to False."""

        # Patch config in the manager module where it's imported
        test_config = Config(VERIFY_COMPATIBILITY=False)
        monkeypatch.setattr("avatars.manager.config", test_config)

        # When should_verify_compatibility is not provided and config is False, should not raise
        try:
            self.manager.authenticate(
                username="username",
                password="password",
            )
        except DeprecationWarning:
            pytest.fail(
                "DeprecationWarning was raised unexpectedly"
                " when config.VERIFY_COMPATIBILITY is False"
            )

    def test_verify_compatibility_parameter_overrides_config(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Verify should_verify_compatibility parameter overrides the config."""

        # Patch config in the manager module where it's imported
        test_config = Config(VERIFY_COMPATIBILITY=True)
        monkeypatch.setattr("avatars.manager.config", test_config)

        # Even though config is True, explicit False parameter should override
        try:
            self.manager.authenticate(
                username="username",
                password="password",
                should_verify_compatibility=False,
            )
        except DeprecationWarning:
            pytest.fail(
                "DeprecationWarning was raised unexpectedly when should_verify_compatibility=False"
            )


class TestManagerInitialization:
    """Tests for Manager config parameter validation."""

    @pytest.fixture
    def minimal_config(self) -> ClientConfig:
        """Provide a minimal ClientConfig for testing."""
        return ClientConfig(
            base_api_url="https://custom.octopize.app/api",
            storage_endpoint_url="https://custom.octopize.app/storage",
        )

    def test_can_create_with_config_only(self, minimal_config: ClientConfig) -> None:
        """Test that Manager can be created with only a ClientConfig object."""
        manager = Manager(config=minimal_config)
        assert manager.auth_client is not None
        assert manager.auth_client.base_url == "https://custom.octopize.app/api"
        assert (
            manager.auth_client.data_uploader.storage_endpoint_url
            == "https://custom.octopize.app/storage"
        )

    def test_can_create_with_base_url_only_deprecated(self) -> None:
        """Test that Manager can be created with only base_url which points to the API."""
        with mock.patch.dict(
            os.environ,
            clear=True,
        ):
            manager = Manager(base_url="https://api.example.com/api")
            assert manager.auth_client is not None
            assert manager.auth_client.base_url == "https://api.example.com/api"
            assert (
                manager.auth_client.data_uploader.storage_endpoint_url
                == "https://api.example.com/storage"
            )

    def test_can_create_with_base_url_only(self) -> None:
        """Test that Manager can be created with only base_url which points to the server."""
        with mock.patch.dict(
            os.environ,
            clear=True,
        ):
            manager = Manager(base_url="https://api.example.com")
            assert manager.auth_client is not None

            # Make sure api url is correctly constructed
            assert manager.auth_client.base_url == "https://api.example.com/api"

            # Make sure storage endpoint URL is correctly constructed
            assert (
                manager.auth_client.data_uploader.storage_endpoint_url
                == "https://api.example.com/storage"
            )

    def test_empty_constructor_uses_defaults(self) -> None:
        """Test that Manager() with no arguments uses default config values."""
        with mock.patch.dict(
            os.environ,
            clear=True,
        ):
            manager = Manager()
            # Manager should create auth_client with default config values
            assert manager.auth_client is not None
            assert manager.auth_client.base_url == "https://www.octopize.app/api"

    def test_empty_constructor_uses_env_vars_base_url(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Verify that Manager() uses AVATAR_BASE_URL env var if set."""
        with mock.patch.dict(
            os.environ,
            clear=True,
        ):
            monkeypatch.setenv("AVATAR_BASE_URL", "https://env.example.com")
            manager = Manager()
            # Manager should create auth_client with config values from env vars
            assert manager.auth_client is not None
            assert manager.auth_client.base_url == "https://env.example.com/api"
            assert (
                manager.auth_client.data_uploader.storage_endpoint_url
                == "https://env.example.com/storage"
            )

    def test_empty_constructor_uses_env_vars_base_api_url(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Verify that Manager() uses AVATAR_BASE_API_URL env var if set.

        This test also ensures that AVATAR_STORAGE_ENDPOINT_URL is NOT inferred
        """
        with mock.patch.dict(
            os.environ,
            clear=True,
        ):
            monkeypatch.setenv("AVATAR_BASE_API_URL", "https://env.example.com/api")
            manager = Manager()
            assert manager.auth_client is not None

            # Verify that BASE_API_URL is set from env var
            assert manager.auth_client.base_url == "https://env.example.com/api"

            # but STORAGE_ENDPOINT_URL is not inferred from BASE_API_URL
            assert (
                manager.auth_client.data_uploader.storage_endpoint_url
                == "https://www.octopize.app/storage"
            )

    def test_empty_constructor_uses_env_vars_storage_endpoint_url(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Verify that Manager() uses AVATAR_STORAGE_ENDPOINT_URL env var if set.

        This test also ensures that AVATAR_BASE_API_URL is NOT inferred
        """
        with mock.patch.dict(
            os.environ,
            clear=True,
        ):
            monkeypatch.setenv("AVATAR_STORAGE_ENDPOINT_URL", "https://env.example.com/storage")
            manager = Manager()
            assert manager.auth_client is not None

            # Verify that BASE_API_URL is set from env var
            assert manager.auth_client.base_url == "https://www.octopize.app/api"

            # but STORAGE_ENDPOINT_URL is not inferred from BASE_API_URL
            assert (
                manager.auth_client.data_uploader.storage_endpoint_url
                == "https://env.example.com/storage"
            )

    def test_no_base_url_and_no_config_but_env_vars_raises_if_conflict(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        with mock.patch.dict(
            os.environ,
            clear=True,
        ):
            monkeypatch.setenv("AVATAR_BASE_API_URL", "https://env.example.com/api")
            monkeypatch.setenv("AVATAR_BASE_URL", "https://env.example.com")

            with pytest.raises(
                ValueError,
                match="Cannot set BASE_URL together with BASE_API_URL or STORAGE_ENDPOINT_URL",
            ):
                Manager()

    def test_config_with_base_url_raises_error(self, minimal_config: ClientConfig) -> None:
        """Test that using config with base_url raises ValueError."""
        with pytest.raises(
            ValueError,
            match="Cannot provide both 'config' and other parameters \\(base_url\\)",
        ):
            Manager(config=minimal_config, base_url="https://other.com")

    def test_config_with_api_key_raises_error(self, minimal_config: ClientConfig) -> None:
        """Test that using config with api_key raises ValueError."""
        with pytest.raises(
            ValueError,
            match="Cannot provide both 'config' and other parameters \\(api_key\\)",
        ):
            Manager(config=minimal_config, api_key="test-key")

    def test_config_with_api_client_raises_error(self, minimal_config: ClientConfig) -> None:
        """Test that using config with api_client raises ValueError."""
        api_client = ApiClient(
            base_url="https://other.com/api",
            verify_auth=False,
        )
        with pytest.raises(
            ValueError,
            match="Cannot provide both 'api_client' and other parameters \\(config\\)",
        ):
            Manager(config=minimal_config, api_client=api_client)


class TestFindSetNamesByDisplayName:
    """Test suite for Manager.find_set_names_by_display_name functionality."""

    def test_returns_matching_set_names(self):
        """Test that matching set_names are returned."""
        set_name1 = uuid4()
        set_name2 = uuid4()
        now = datetime.now(UTC)

        jobs = [
            create_fake_job(
                name="job1",
                set_name=set_name1,
                created_at=now - timedelta(days=1),
                display_name="test_dataset",
            ),
            create_fake_job(
                name="job2",
                set_name=set_name2,
                created_at=now - timedelta(days=5),
                display_name="test_dataset",
            ),
        ]
        fake_client = FakeApiClient()
        for job in jobs:
            fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        results = manager.find_ids_by_name("test_dataset")

        assert len(results) == 2
        assert results[0][0] == str(set_name1)  # Newest first
        assert results[1][0] == str(set_name2)

    def test_returns_empty_list_when_no_matches(self):
        """Test that empty list is returned when no display_name matches."""
        jobs = [
            create_fake_job(display_name="other_dataset"),
        ]
        fake_client = FakeApiClient()
        for job in jobs:
            fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        results = manager.find_ids_by_name("test_dataset")

        assert results == []

    def test_case_sensitive_matching(self):
        """Test that display_name matching is case-sensitive."""
        jobs = [
            create_fake_job(display_name="Test_Dataset"),
            create_fake_job(display_name="test_dataset"),
        ]
        fake_client = FakeApiClient()
        for job in jobs:
            fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        results = manager.find_ids_by_name("test_dataset")

        assert len(results) == 1
        assert results[0][1][0].display_name == "test_dataset"

    def test_excludes_advice_jobs(self):
        """Test that advice jobs are excluded from results."""
        set_name_standard = uuid4()
        set_name_advice = uuid4()

        jobs = [
            create_fake_job(
                name="standard",
                set_name=set_name_standard,
                kind=JobKind.standard,
                display_name="test_dataset",
            ),
            create_fake_job(
                name="advice",
                set_name=set_name_advice,
                kind=JobKind.advice,
                display_name="test_dataset",
            ),
        ]
        fake_client = FakeApiClient()
        for job in jobs:
            fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        results = manager.find_ids_by_name("test_dataset")

        assert len(results) == 1
        assert results[0][0] == str(set_name_standard)

    def test_sorts_by_newest_first(self):
        """Test that results are sorted by creation time (newest first)."""
        now = datetime.now(UTC)
        set_name_oldest = uuid4()
        set_name_middle = uuid4()
        set_name_newest = uuid4()

        jobs = [
            create_fake_job(
                name="middle",
                set_name=set_name_middle,
                created_at=now - timedelta(days=5),
                display_name="test_dataset",
            ),
            create_fake_job(
                name="oldest",
                set_name=set_name_oldest,
                created_at=now - timedelta(days=10),
                display_name="test_dataset",
            ),
            create_fake_job(
                name="newest",
                set_name=set_name_newest,
                created_at=now - timedelta(days=1),
                display_name="test_dataset",
            ),
        ]
        fake_client = FakeApiClient()
        for job in jobs:
            fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        results = manager.find_ids_by_name("test_dataset")

        assert len(results) == 3
        assert results[0][0] == str(set_name_newest)
        assert results[1][0] == str(set_name_middle)
        assert results[2][0] == str(set_name_oldest)

    def test_handles_multiple_jobs_same_set_name(self):
        """Test that multiple jobs with the same set_name are not deduplicated."""
        set_name = uuid4()
        now = datetime.now(UTC)

        jobs = [
            create_fake_job(
                name="standard",
                set_name=set_name,
                kind=JobKind.standard,
                created_at=now - timedelta(days=1),
                display_name="test_dataset",
            ),
            create_fake_job(
                name="privacy_metrics",
                set_name=set_name,
                kind=JobKind.privacy_metrics,
                created_at=now - timedelta(days=2),
                display_name="test_dataset",
            ),
            create_fake_job(
                name="signal_metrics",
                set_name=set_name,
                kind=JobKind.signal_metrics,
                created_at=now - timedelta(days=3),
                display_name="test_dataset",
            ),
        ]
        fake_client = FakeApiClient()
        for job in jobs:
            fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        results = manager.find_ids_by_name("test_dataset")

        # Only one entry for the set_name should be returned
        assert len(results) == 1

    def test_integration_with_create_runner_from_id(self):
        """Test the workflow: find set_names -> load runner by id."""
        set_name = uuid4()
        job = create_fake_job(set_name=set_name, display_name="test_dataset")
        fake_client = FakeApiClient()
        fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        found_set_names = manager.find_ids_by_name("test_dataset")
        assert len(found_set_names) == 1

        runner = manager.create_runner_from_id(found_set_names[0][0])

        assert runner is not None
        assert runner.set_name == str(set_name)
        assert runner.display_name == "test_dataset-v1"


class TestCreateRunnerFromId:
    def test_accepts_uuid_string(self):
        """Test that a valid UUID string is accepted."""
        set_name = uuid4()
        fake_client = FakeApiClient()
        manager = Manager(api_client=fake_client)

        runner = manager.create_runner_from_id(str(set_name))

        assert runner is not None
        assert runner.set_name == str(set_name)

    def test_accepts_uuid_object(self):
        """Test that a UUID object is accepted and converted to string."""
        set_name = uuid4()
        fake_client = FakeApiClient()
        manager = Manager(api_client=fake_client)

        runner = manager.create_runner_from_id(set_name)

        assert runner is not None
        assert runner.set_name == str(set_name)

    def test_rejects_invalid_uuid_string(self):
        """Test that an invalid UUID string raises ValueError."""
        fake_client = FakeApiClient()
        manager = Manager(api_client=fake_client)

        with pytest.raises(ValueError, match="Invalid set_name format"):
            manager.create_runner_from_id("not-a-uuid")

    def test_rejects_none(self):
        """Test that None raises TypeError."""
        fake_client = FakeApiClient()
        manager = Manager(api_client=fake_client)

        with pytest.raises(TypeError, match="str or UUID"):
            manager.create_runner_from_id(None)  # type: ignore[arg-type]

    def test_populates_runner_state(self):
        """Test that runner.config, jobs.config, and jobs.set_name are set correctly."""
        set_name = uuid4()
        fake_client = FakeApiClient()
        manager = Manager(api_client=fake_client)

        runner = manager.create_runner_from_id(str(set_name))

        assert runner.set_name == str(set_name)
        assert runner.jobs.set_name == str(set_name)
        assert runner.config is not None
        assert runner.jobs.config is not None

    def test_create_runner_from_id_increments_display_name(self):
        """Loading a runner via create_runner_from_id gives an incremented display name."""
        set_name = uuid4()
        job = create_fake_job(set_name=set_name, display_name="test_dataset")
        fake_client = FakeApiClient()
        fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        runner = manager.create_runner_from_id(str(set_name))

        assert runner.display_name == "test_dataset-v1"


class TestCreateRunnerFromSetName:
    def test_raises_value_error_when_no_match(self):
        """Test that ValueError is raised when no jobs match the display name."""
        fake_client = FakeApiClient()
        manager = Manager(api_client=fake_client)

        with pytest.raises(ValueError, match="No jobs found for display name 'unknown'"):
            manager.create_runner_from_name("unknown")

    def test_returns_runner_for_single_match(self):
        """Test that a runner is returned when exactly one set_name matches."""
        set_name = uuid4()
        job = create_fake_job(set_name=set_name, display_name="my_dataset", kind=JobKind.standard)
        fake_client = FakeApiClient()
        fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        runner = manager.create_runner_from_name("my_dataset")

        assert runner is not None
        assert runner.set_name == str(set_name)

    def test_picks_most_recent_when_multiple_set_names(self):
        """Test that the most recent set_name is selected when multiple match."""
        now = datetime.now(UTC)
        set_name_old = uuid4()
        set_name_new = uuid4()

        jobs = [
            create_fake_job(
                set_name=set_name_old,
                display_name="my_dataset",
                created_at=now - timedelta(days=5),
            ),
            create_fake_job(
                set_name=set_name_new,
                display_name="my_dataset",
                created_at=now - timedelta(days=1),
            ),
        ]
        fake_client = FakeApiClient()
        for job in jobs:
            fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        runner = manager.create_runner_from_name("my_dataset")

        assert runner.set_name == str(set_name_new)

    def test_create_runner_from_name_increments_display_name(self):
        """Loading a runner via create_runner_from_name gives an incremented display name."""
        set_name = uuid4()
        job = create_fake_job(set_name=set_name, display_name="test_dataset")
        fake_client = FakeApiClient()
        fake_client.jobs.add_job(job)
        manager = Manager(api_client=fake_client)

        runner = manager.create_runner_from_name("test_dataset")

        assert runner.display_name == "test_dataset-v1"


class TestIncrementDisplayNameVersion:
    def test_appends_v1_when_no_version(self):
        assert _increment_display_name_version("my_dataset") == "my_dataset-v1"

    def test_increments_existing_version(self):
        assert _increment_display_name_version("my_dataset-v1") == "my_dataset-v2"

    def test_increments_multi_digit_version(self):
        assert _increment_display_name_version("my_dataset-v9") == "my_dataset-v10"

    def test_does_not_match_mid_string_version(self):
        """A -vN that is not at the end is not treated as a version."""
        assert _increment_display_name_version("my-v2-dataset") == "my-v2-dataset-v1"

    def test_create_runner_from_yaml_increments_display_name(self):
        """Loading a runner via create_runner_from_yaml gives an incremented display name."""
        fake_client = FakeApiClient()
        manager = Manager(api_client=fake_client)

        runner = manager.create_runner_from_yaml("fixtures/yaml_from_web.yaml", "iris")

        assert runner.display_name == "iris-v1"


class TestManagerDeleteJobs:
    """Tests for Manager.delete_job, Manager.delete_job_by_id and Manager.delete_jobs."""

    def setup_method(self) -> None:
        self.fake_api_client = FakeApiClient()
        self.manager = Manager(api_client=self.fake_api_client)  # type: ignore[arg-type]

    def test_delete_job_by_display_name(self) -> None:
        set_name = uuid4()
        job = create_fake_job(set_name=set_name, display_name="my_run", kind=JobKind.standard)
        self.fake_api_client.jobs.add_job(job)

        result = self.manager.delete_job("my_run")

        assert isinstance(result, BulkDeleteResponse)
        assert len(result.deleted_jobs) == 1
        assert result.deleted_jobs[0].name == job.name

    def test_delete_job_raises_when_no_match(self) -> None:
        with pytest.raises(ValueError, match="No jobs found for display name 'nonexistent'"):
            self.manager.delete_job("nonexistent")

    def test_delete_job_raises_when_multiple_sets(self) -> None:
        set_name1 = uuid4()
        set_name2 = uuid4()
        self.fake_api_client.jobs.add_job(
            create_fake_job(
                name="job-set1",
                set_name=set_name1,
                display_name="shared_name",
                kind=JobKind.standard,
            )
        )
        self.fake_api_client.jobs.add_job(
            create_fake_job(
                name="job-set2",
                set_name=set_name2,
                display_name="shared_name",
                kind=JobKind.standard,
            )
        )

        with pytest.raises(ValueError, match="Multiple runs found"):
            self.manager.delete_job("shared_name")

    def test_delete_job_raises_error_message_contains_setnames(self) -> None:
        set_name1 = uuid4()
        set_name2 = uuid4()
        self.fake_api_client.jobs.add_job(
            create_fake_job(
                name="job-a", set_name=set_name1, display_name="shared_name", kind=JobKind.standard
            )
        )
        self.fake_api_client.jobs.add_job(
            create_fake_job(
                name="job-b", set_name=set_name2, display_name="shared_name", kind=JobKind.standard
            )
        )

        with pytest.raises(ValueError) as exc_info:
            self.manager.delete_job("shared_name")

        msg = str(exc_info.value)
        assert str(set_name1) in msg or str(set_name2) in msg
        assert "delete_job_by_id" in msg

    def test_delete_job_by_id(self) -> None:
        set_name = uuid4()
        job1 = create_fake_job(
            name="job1", set_name=set_name, kind=JobKind.standard, display_name="run1"
        )
        job2 = create_fake_job(
            name="job2", set_name=set_name, kind=JobKind.privacy_metrics, display_name="run1"
        )
        self.fake_api_client.jobs.add_job(job1)
        self.fake_api_client.jobs.add_job(job2)

        result = self.manager.delete_job_by_id(set_name)

        assert isinstance(result, BulkDeleteResponse)
        assert len(result.deleted_jobs) == 2
        assert result.failed_jobs == []

    def test_delete_job_by_id_no_matching_jobs(self) -> None:
        result = self.manager.delete_job_by_id(uuid4())
        assert isinstance(result, BulkDeleteResponse)
        assert result.deleted_jobs == []
        assert result.failed_jobs == []

    def test_delete_jobs_single_batch(self) -> None:
        job_names = [
            self.fake_api_client.jobs.create_job(
                JobCreateRequest(set_name=uuid4(), parameters_name=f"params-{i}")
            ).name
            for i in range(50)
        ]
        result = self.manager.delete_jobs(job_names)
        assert isinstance(result, BulkDeleteResponse)
        assert len(result.deleted_jobs) == 50
        assert result.failed_jobs == []

    def test_delete_jobs_not_found(self) -> None:
        result = self.manager.delete_jobs(["nonexistent-1", "nonexistent-2"])
        assert isinstance(result, BulkDeleteResponse)
        assert result.deleted_jobs == []
        assert result.failed_jobs == ["nonexistent-1", "nonexistent-2"]

    def test_delete_jobs_multiple_batches(self) -> None:
        """Verify that >100 job names are split into multiple 100-job API calls."""
        job_names = [
            self.fake_api_client.jobs.create_job(
                JobCreateRequest(set_name=uuid4(), parameters_name=f"params-{i}")
            ).name
            for i in range(250)
        ]
        result = self.manager.delete_jobs(job_names)

        # Should have made 3 calls: 100 + 100 + 50
        calls = self.fake_api_client.jobs.bulk_delete_calls
        assert len(calls) == 3
        assert calls[0] == BulkDeleteRequest(job_names=job_names[:100])
        assert calls[1] == BulkDeleteRequest(job_names=job_names[100:200])
        assert calls[2] == BulkDeleteRequest(job_names=job_names[200:])
        assert len(result.deleted_jobs) == 250
        assert result.failed_jobs == []

    def test_delete_jobs_empty_list(self) -> None:
        result = self.manager.delete_jobs([])
        assert isinstance(result, BulkDeleteResponse)
        assert result.deleted_jobs == []
        assert result.failed_jobs == []



import re
from pathlib import Path
from typing import Any, TypeGuard

import pandas as pd
from IPython.display import HTML
from pydantic import BaseModel

from avatars.constants import (
    RESULTS_TO_STORE,
    PlotKind,
    Results,
    TypeResults,
    mapping_result_to_file_name,
)


class ResultsOrganizer(BaseModel):
    model_config = {"arbitrary_types_allowed": True}

    shuffled: dict[str, pd.DataFrame] = {}
    sensitive_unshuffled: dict[str, pd.DataFrame] = {}
    privacy_metrics: dict[str, list[dict[str, Any]]] = {}
    signal_metrics: dict[str, list[dict[str, Any]]] = {}
    original_projections: dict[str, pd.DataFrame] = {}
    avatars_projections: dict[str, pd.DataFrame] = {}
    figures: dict[str, dict[str, list[HTML]]] = {}
    advice: dict[str, dict[str, Any]] = {}
    run_metadata: dict[str, Any] = {}

    def get_results(self, table_name: str, result_name: Results, job_name: str) -> TypeResults:
        table_name = self.unescape_table_name(table_name)
        match result_name:
            case Results.SHUFFLED:
                return self.shuffled.get(table_name)
            case Results.UNSHUFFLED:
                return self.sensitive_unshuffled.get(table_name)
            case Results.PRIVACY_METRICS:
                return self.privacy_metrics.get(table_name)
            case Results.SIGNAL_METRICS:
                return self.signal_metrics.get(table_name)
            case Results.PROJECTIONS_ORIGINAL:
                return self.original_projections.get(table_name)
            case Results.PROJECTIONS_AVATARS:
                return self.avatars_projections.get(table_name)
            case Results.FIGURES:
                return self.figures.get(table_name)
            case Results.ADVICE:
                return self.advice.get(table_name, {})
            case Results.METADATA:
                return self.run_metadata.get(job_name, {})
            case _:
                raise ValueError(f"Expected result_name to be one of {RESULTS_TO_STORE}")

    def set_results(
        self,
        result_name: Results,
        result: TypeResults,
        table_name: str,
        metadata: dict[str, Any] | None = None,
    ):
        table_name = self.unescape_table_name(table_name)
        match result_name:
            case Results.SHUFFLED:
                if self._is_dataframe(result):
                    self.shuffled[table_name] = result
            case Results.UNSHUFFLED:
                if self._is_dataframe(result):
                    self.sensitive_unshuffled[table_name] = result
            case Results.PRIVACY_METRICS:
                if self._is_list_of_dict(result):
                    self.privacy_metrics.setdefault(table_name, [])
                    self.privacy_metrics[table_name].append(result[0])
            case Results.SIGNAL_METRICS:
                if self._is_list_of_dict(result):
                    self.signal_metrics.setdefault(table_name, [])
                    self.signal_metrics[table_name].append(result[0])
            case Results.PROJECTIONS_ORIGINAL:
                if self._is_dataframe(result):
                    self.original_projections[table_name] = result
            case Results.PROJECTIONS_AVATARS:
                if self._is_dataframe(result):
                    self.avatars_projections[table_name] = result
            case Results.FIGURES:
                if self._is_html(result) and metadata is not None:
                    self.figures.setdefault(table_name, {})
                    if metadata["kind"] in PlotKind:
                        self.figures[table_name].setdefault(PlotKind(metadata["kind"]), []).append(
                            result
                        )
            case Results.ADVICE:
                if isinstance(result, list):
                    for item in result:
                        if item["table_name"] == table_name:
                            self.advice[table_name] = item["advice"]
                            break
            case Results.METADATA:
                if isinstance(result, dict) and metadata is not None:
                    self.run_metadata[metadata["kind"]] = result
            case _:
                raise ValueError(f"Expected result_name to be one of {RESULTS_TO_STORE}")

    def _is_dataframe(self, result: TypeResults) -> TypeGuard[pd.DataFrame]:
        """Validate that the result is a DataFrame."""
        if not isinstance(result, pd.DataFrame):
            raise ValueError("Expected result to be a DataFrame")
        return True

    def _is_list_of_dict(self, result: TypeResults) -> TypeGuard[list[dict[str, Any]]]:
        """Validate that the result is a list of dict."""
        if not isinstance(result, list) or not all(isinstance(r, dict) for r in result):
            raise ValueError("Expected result to be a list of dicts")
        return True

    def _is_html(self, result: TypeResults) -> TypeGuard[HTML]:
        """Validate that the result is HTML."""
        if not isinstance(result, HTML):
            raise ValueError("Expected result to be an HTML object for FIGURES")
        return True

    def get_table_name(
        self,
        result_name: Results,
        url: str,
        result: TypeResults,
        metadata: TypeResults | None = None,
    ) -> str | None:
        """Get the table name from the result based on the result type and URL.
        This method extracts the table name in the URL for tables,
        or from the metadata for other result types."""
        table_name = None
        match result_name:
            case (
                Results.SHUFFLED
                | Results.UNSHUFFLED
                | Results.PROJECTIONS_ORIGINAL
                | Results.PROJECTIONS_AVATARS
            ):
                result_filename = mapping_result_to_file_name[result_name]
                table_name = Path(url).stem.split(f".{result_filename}")[0]
            case Results.SIGNAL_METRICS | Results.PRIVACY_METRICS:
                if isinstance(result, list) and result:
                    table_name = result[0]["metadata"]["table_name"]
                else:
                    raise ValueError(
                        "Expected result to be a list with at least one item for METRICS"
                    )
            case Results.FIGURES:
                if metadata and isinstance(metadata, dict):
                    table_name = metadata.get("table_name", "")
            case Results.ADVICE:
                if isinstance(result, list) and result:
                    table_name = result[0].get("table_name", "")
        return table_name

    def unescape_table_name(self, table_name: str) -> str:
        """Encode the table name to be used in URLs."""
        reversed_replacements = {
            "_dot_": ".",
            "_slash_": "/",
            "_ampersand_": "&",
            "_plus_": "+",
        }
        pattern = re.compile("|".join(re.escape(k) for k in reversed_replacements.keys()))
        return pattern.sub(lambda match: reversed_replacements[match.group(0)], table_name)

    def find_figure_metadata(
        self, figures_metadatas: list[dict[str, Any]], url: str
    ) -> dict[str, Any] | None:
        for metadata in figures_metadatas:
            if metadata.get("filename") == Path(url).name:
                figure_metadata = metadata
                return figure_metadata
        return None

    def is_results_empty(self) -> bool:
        """Check if all results are empty."""
        return not any(
            [
                self.shuffled,
                self.sensitive_unshuffled,
                self.privacy_metrics,
                self.signal_metrics,
                self.original_projections,
                self.avatars_projections,
                self.figures,
                self.run_metadata,
            ]
        )



"""Job launcher for orchestrating avatarization jobs with proper dependencies and validation."""

from typing import Any
from uuid import UUID

from avatar_yaml import Config
from avatar_yaml.models.advice import AdviceType

from avatars.client import ApiClient
from avatars.models import JobCreateRequest, JobCreateResponse, JobKind, JobResponse


class JobLauncher:
    """Orchestrates job creation with dependency management and prerequisite validation."""

    def __init__(self, client: ApiClient, config: Config) -> None:
        """Initialize the JobLauncher.

        Parameters
        ----------
        client
            The API client to interact with the backend.
        config
            The avatarization configuration.
        """
        self.client = client
        self.config = config
        self.set_name: str = ""
        self.jobs: dict[str, JobCreateResponse] = {}

    def register_existing_job(
        self, parameters_name: str, job_name: str, job_location: str = ""
    ) -> None:
        """Register a job that was created previously (for reconstructed runners).

        This allows reconstructed runners to access results from jobs that were
        launched in a previous session.

        Parameters
        ----------
        parameters_name
            The name of the parameters associated with the job.
        job_name
            The name/parameters_name of the job to register.
        job_location
            Optional location URL for the job. Defaults to /jobs/{job_name}.
        """
        self.jobs[parameters_name] = JobCreateResponse(
            name=job_name,
            Location=job_location or f"/jobs/{job_name}",
        )

    def launch_job(
        self,
        job_kind: JobKind,
        set_name: str,
    ) -> None:
        """Launch a job with automatic dependency resolution and validation.

        Parameters
        ----------
        job_kind
            The kind of job to launch.
        set_name
            The set name (UUID) for the job.
        Returns
        -------
        JobCreateResponse
            The created job response.

        Raises
        ------
        ValueError
            If required prerequisites are not met for the job.
        """
        self.set_name = set_name
        self._validate_job_prerequisites(job_kind)
        depends_on = self._get_job_dependencies(job_kind)
        self._create_jobs(job_kind=job_kind, depends_on=depends_on)

    def _create_jobs(self, job_kind: JobKind, depends_on) -> None:
        """Create all jobs based on the configuration and dependencies."""
        parameters_name = self.get_parameters_name(job_kind)
        created_job = self._create_job(parameters_name=parameters_name, depends_on=depends_on)
        self.jobs[parameters_name] = created_job

        # For report jobs, also create the PIA report
        if job_kind == JobKind.report:
            pia_parameters_name = self.get_parameters_name(job_kind, pia_report=True)
            pia_job = self._create_job(parameters_name=pia_parameters_name, depends_on=depends_on)
            self.jobs[pia_parameters_name] = pia_job

    def get_parameters_name(self, job_kind: JobKind, pia_report: bool = False) -> str:
        """Get the parameters name for a given job kind.

        Parameters
        ----------
        job_kind
            The kind of job to get parameters name for.

        Returns
        -------
        str
            The parameters name.
        """
        match job_kind:
            case JobKind.report:
                if pia_report:
                    return "report_pia"
                return "report"
            case JobKind.advice:
                return self.config.get_parameters_advice_name(
                    name="advice", advisor_type=[AdviceType.PARAMETERS]
                )
            case str():
                return job_kind
            case _:
                return job_kind.value

    def _validate_job_prerequisites(self, job_kind: JobKind) -> None:
        """Validate that prerequisites are met before launching a job.

        Parameters
        ----------
        job_kind
            The kind of job to validate.

        Raises
        ------
        ValueError
            If required prerequisites are not met.
        """
        match job_kind:
            case JobKind.standard:
                avatarization_open_dp = getattr(self.config, "avatarization_open_dp", None)
                avatarization_fast_dp = getattr(self.config, "avatarization_fast_dp", None)
                if (
                    not self.config.avatarization
                    and not avatarization_open_dp
                    and not avatarization_fast_dp
                ):
                    raise ValueError(
                        "Expected k or epsilon to be set to run an avatarization job, "
                        "You have to set a k or epsilon parameter using `runner.set_parameter()`."
                    )

            case JobKind.signal_metrics | JobKind.privacy_metrics:
                # If standard job wasn't run, avatar tables must be provided
                if JobKind.standard.value not in self.jobs:
                    for avatar_table in self.config.avatar_tables.values():
                        if (
                            avatar_table.avatars_data is None
                            or avatar_table.avatars_data.file is None
                        ):
                            raise ValueError(
                                "Expected Avatar tables to be set to run signal/privacy metrics "
                                "job. You have to set avatar_data using `runner.add_table()`."
                            )

            case JobKind.report:
                if (
                    JobKind.privacy_metrics.value not in self.jobs
                    or JobKind.signal_metrics.value not in self.jobs
                ):
                    raise ValueError(
                        "Expected Privacy and Signal jobs to be created before running report."
                    )
            case _:
                pass

    def _get_job_dependencies(self, job_kind: JobKind) -> list[str]:
        """Get the list of job dependencies for a given job kind.

        Parameters
        ----------
        job_kind
            The kind of job to get dependencies for.

        Returns
        -------
        list[str]
            List of job locations that this job depends on.
        """
        match job_kind:
            case JobKind.signal_metrics | JobKind.privacy_metrics:
                avatarization_job_name = self.get_parameters_name(JobKind.standard)
                if avatarization_job_name in self.jobs:
                    return [self.jobs[avatarization_job_name].Location]

            case JobKind.report:
                privacy_metrics_job_name = self.get_parameters_name(JobKind.privacy_metrics)
                signal_metrics_job_name = self.get_parameters_name(JobKind.signal_metrics)
                return [
                    self.jobs[privacy_metrics_job_name].Location,
                    self.jobs[signal_metrics_job_name].Location,
                ]

            case _:
                pass
        return []

    def _create_job(
        self,
        parameters_name: str,
        depends_on: list[str] | None = None,
    ) -> JobCreateResponse:
        """Create a job via the API.

        Parameters
        ----------
        parameters_name
            The name of the parameters to use for the job.
        depends_on
            List of job locations this job depends on.

        Returns
        -------
        JobCreateResponse
            The created job response from the API.
        """
        if depends_on is None:
            depends_on = []

        request = JobCreateRequest(
            set_name=UUID(self.set_name), parameters_name=parameters_name, depends_on=depends_on
        )

        kwargs: dict[str, Any] = {
            "method": "post",
            "url": "/jobs",
            "json_data": request,
        }
        created_job = JobCreateResponse(**self.client.send_request(**kwargs))
        return created_job

    def get_job_response(self, job_name: JobKind | str) -> JobCreateResponse:
        """Get a job by name.

        Parameters
        ----------
        job_name
            The name of the job to get (JobKind enum or string like "report_pia").

        Returns
        -------
        JobCreateResponse
            The job object.

        Raises
        ------
        ValueError
            If the job has not been created yet.
        """
        if isinstance(job_name, JobKind):
            job_name = self.get_parameters_name(job_name)
        if job_name not in self.jobs:
            raise ValueError(f"Expected job '{job_name}' to be created. Try running it first.")
        return self.jobs[job_name]

    def get_job_id(self, job_name: JobKind | str) -> str:
        """Get a job by name.

        Parameters
        ----------
        job_name
            The name of the job to get (JobKind enum or string like "report_pia").

        Returns
        -------
        str
            The job ID.

        Raises
        ------
        ValueError
            If the job has not been created yet.
        """
        job_create_response = self.get_job_response(job_name)
        return job_create_response.name

    def has_job(self, job_name: JobKind | str) -> bool:
        """Check if a job has been created.

        Parameters
        ----------
        job_name
            The name of the job to check.

        Returns
        -------
        bool
            True if the job exists, False otherwise.
        """
        if isinstance(job_name, JobKind):
            job_name = self.get_parameters_name(job_name)
        return job_name in self.jobs

    def get_launched_jobs(self) -> list[str]:
        """Get a list of all jobs that have been launched.

        Returns
        -------
        list[str]
            List of job kinds that have been launched.
        """
        return list(self.jobs.keys())

    def get_job_status(self, job_name: JobKind | str) -> JobResponse:
        """Get the full job status from the API.

        Parameters
        ----------
        job_name
            The name of the job to get status for (JobKind enum or string like "report_pia").

        Returns
        -------
        JobResponse
            The job status response from the API.

        Raises
        ------
        ValueError
            If the job has not been created yet.
        """
        job_id = self.get_job_id(job_name)
        return self.client.jobs.get_job_status(job_id)
